"""
Read/use config/pyproject.toml file (:mod:`~pyproject2conda.config`)
====================================================================
"""

from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from pyproject2conda.utils import filename_from_template, get_in

if TYPE_CHECKING:
    from typing import Any, Iterator, Sequence

    from ._typing_compat import Self


# * Utilities


class Config:
    """Class to parse toml file with [tool.pyproject2conda] section"""

    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data

    def get_in(self, *keys: str, default: Any = None) -> Any:
        """Utility to extract from nested dict."""
        keys = keys
        return get_in(keys=keys, nested_dict=self.data, default=default)

    @cached_property
    def overrides(self) -> list[Any]:
        """All overrides sections from `[[tool.pyproject2conda.overrides]]`"""
        out: list[dict[str, Any]] = []
        for x in self.get_in("overrides", default=[]):
            if "envs" not in x:
                raise ValueError("must specify env in overrides")
            out.append(x)
        return out

    @property
    def envs(self) -> dict[str, Any]:
        """All environments"""
        return self.get_in("envs", default={})  # type: ignore

    def _get_override(self, env: str) -> dict[str, Any]:
        out: dict[str, Any] = {}
        for override in self.overrides:
            if env in override["envs"]:
                out.update(**override)
        if "envs" in out:
            del out["envs"]
        return out

    # def _get_env(self, env_name: str) -> dict[str, Any]:
    #     env = self.get_in("envs", env_name)
    #     env.update(**self._get_override(env_name))
    #     return env  # type: ignore

    def _get_value(
        self,
        key: str,
        env_name: str | None = None,
        inherit: bool = True,
        as_list: bool = False,
        default: Any = None,
    ) -> Any:
        """Get a value from thing"""

        if env_name is None:
            value = self.get_in(key, default=None)

        else:
            # try to get from env definition
            assert env_name in self.data["envs"], f"env {env_name} not in config"

            value = self.get_in("envs", env_name, key, default=None)

            if inherit:
                # try to get from override
                if value is None:
                    value = self._get_override(env_name).get(key, None)

                # finally, try to get from top level
                if value is None:
                    value = self.get_in(key, default=None)

        if value is None:
            if callable(default):
                value = default()
            else:
                value = default

        if value is not None and as_list:
            if not isinstance(value, list):
                value = [value]

        return value  # pyright: ignore

    def channels(
        self, env_name: str | None = None, inherit: bool = True
    ) -> list[str] | None:
        """Channels getter"""
        return self._get_value(  # type: ignore
            key="channels", env_name=env_name, inherit=inherit, as_list=True
        )

    def python(
        self, env_name: str | None = None, inherit: bool = True, default: Any = list
    ) -> list[str]:
        """Python getter"""

        # if callable(default):
        #     default = default()

        return self._get_value(  # type: ignore
            key="python",
            env_name=env_name,
            inherit=inherit,
            as_list=True,
            default=default,
        )

    def extras(self, env_name: str) -> list[str]:
        """
        Extras getter

        * If value is `True` (default), then return [env_name]
        * If value is `False`, return []
        * else return list of extras
        """

        val = self._get_value(
            key="extras",
            env_name=env_name,
            inherit=False,
            # as_list=True,
            default=env_name,
        )

        if isinstance(val, bool):
            if val:
                return [env_name]
            else:
                return []

        elif not isinstance(val, list):
            val = [val]

        return val  # type: ignore

    def output(self, env_name: str | None = None) -> str | None:
        """Output getter"""
        return self._get_value(key="output", env_name=env_name, inherit=False)  # type: ignore

    def sort(
        self, env_name: str | None = None, inherit: bool = True, default: bool = True
    ) -> bool:
        """Sort getter"""
        return self._get_value(  # type: ignore
            key="sort", env_name=env_name, inherit=inherit, default=default
        )

    # def inherit(self, env_name: str, default: bool = True) -> bool:
    #     return self._get_value(  # type: ignore
    #         key="inherit", env_name=env_name, inherit=True, default=default
    #     )

    def base(self, env_name: str, default: bool = True) -> bool:
        """Base getter."""
        return self._get_value(key="base", env_name=env_name, default=default)  # type: ignore

    def name(self, env_name: str) -> bool:
        return self._get_value(key="name", env_name=env_name)  # type: ignore

    def header(self, env_name: str) -> bool:
        """Header getter"""
        return self._get_value(key="header", env_name=env_name)  # type: ignore

    def style(self, env_name: str | None = None, default: str = "yaml") -> str:
        """Style getter.  One of `yaml`, `requirements`"""
        out = self._get_value(
            key="style", env_name=env_name, default=default, as_list=True
        )
        for k in out:
            assert k in ["yaml", "requirements", "conda-requirements", "json"]
        return out  # type: ignore

    def python_include(self, env_name: str | None = None) -> str | None:
        """Flag python_include"""
        return self._get_value(key="python_include", env_name=env_name)  # type: ignore

    def python_version(self, env_name: str | None = None) -> str | None:
        """Flag python_version"""
        return self._get_value(key="python_version", env_name=env_name)  # type: ignore

    def overwrite(self, env_name: str | None = None, default: str = "check") -> str:
        """Flag overwrite"""
        return self._get_value(key="overwrite", env_name=env_name, default=default)  # type: ignore

    def verbose(
        self, env_name: str | None = None, default: int | None = None
    ) -> int | None:
        """Flag verbose"""
        return self._get_value(key="verbose", env_name=env_name, default=default)  # type: ignore

    def template(self, env_name: str, default: str = "{env}") -> str:
        """Flag for template"""
        return self._get_value(key="template", env_name=env_name, default=default)  # type: ignore

    def template_python(self, env_name: str, default: str = "py{py}-{env}") -> str:
        """Flag for template_python."""
        return self._get_value(  # type: ignore
            key="template_python", env_name=env_name, default=default
        )

    def deps(self, env_name: str, default: Any = None) -> list[str]:
        return self._get_value(  # type: ignore
            key="deps",
            env_name=env_name,
            default=default,
        )

    def reqs(self, env_name: str, default: Any = None) -> list[str]:
        return self._get_value(  # type: ignore
            key="reqs",
            env_name=env_name,
            default=default,
        )

    def user_config(self, env_name: str | None = None) -> str | None:  # pyright: ignore
        """Flag user_config"""
        return self._get_value(key="user_config", default=None)  # type: ignore

    def allow_empty(self, env_name: str | None = None, default: bool = False) -> bool:
        return self._get_value(  # type: ignore
            key="allow_empty", env_name=env_name, default=default
        )

    def remove_whitespace(
        self, env_name: str | None = None, default: bool = True
    ) -> bool:
        return self._get_value(  # type: ignore
            key="remove_whitespace",
            env_name=env_name,
            default=default,
        )

    def assign_user_config(self, user: Self) -> Self:
        """Assign user_config to self."""
        from copy import deepcopy

        data = deepcopy(self.data)

        # get user envs

        if "envs" not in data:
            data["envs"] = {}

        if "overrides" not in data:
            data["overrides"] = []

        for key in ["envs", "overrides"]:
            u = user.get_in(key)
            if u is not None:
                d = data[key]
                if isinstance(d, list):
                    assert isinstance(u, list)
                    d.extend(u)  # pyright: ignore
                elif isinstance(d, dict):
                    assert isinstance(u, dict)
                    d.update(**u)  # pyright: ignore

        return type(self)(data)

    def _get_output_and_templates(
        self, env_name: str, **defaults: Any
    ) -> list[str | None]:
        return [
            defaults.get(k, getattr(self, k)(env_name))
            for k in ["output", "template", "template_python"]
        ]

    def _iter_yaml(
        self, env_name: str, **defaults: Any
    ) -> Iterator[tuple[str, dict[str, Any]]]:
        pythons = self.python(env_name)
        output, template, template_python = self._get_output_and_templates(
            env_name, **defaults
        )

        keys = [
            "extras",
            "sort",
            "base",
            "header",
            "overwrite",
            "verbose",
            "reqs",
            "deps",
            "name",
            "channels",
            "allow_empty",
            "remove_whitespace",
        ]

        data = {k: defaults.get(k, getattr(self, k)(env_name)) for k in keys}

        if not pythons:
            if output is None:
                output = filename_from_template(
                    template=template, env_name=env_name, ext="yaml"
                )

            if python_include := self.python_include(env_name):
                data.update(python_include=python_include)

            if python_version := self.python_version(env_name):
                data.update(python_version=python_version)

            data.update(output=output)
            yield ("yaml", data)

        else:
            for python in pythons:
                output = filename_from_template(
                    template=template_python,
                    python=python,
                    env_name=env_name,
                    ext="yaml",
                )
                data.update(python=python, output=output)
                yield ("yaml", data)

    def _iter_reqs(
        self, env_name: str, **defaults: Any
    ) -> Iterator[tuple[str, dict[str, Any]]]:
        keys = [
            "extras",
            "sort",
            "base",
            "header",
            "overwrite",
            "verbose",
            "reqs",
            "allow_empty",
            "remove_whitespace",
        ]

        output, template, _ = self._get_output_and_templates(env_name, **defaults)

        data = {k: defaults.get(k, getattr(self, k)(env_name)) for k in keys}

        output = self.output(env_name)
        if not output:
            output = filename_from_template(
                template=template, env_name=env_name, ext="txt"
            )
        data.update(output=output)

        yield ("requirements", data)

    def iter(
        self, envs: Sequence[str] | None = None, **defaults: Any
    ) -> Iterator[tuple[str, dict[str, Any]]]:
        """Iterate over configs"""

        # filter defaults.  Only include values of not None:
        defaults = {k: v for k, v in defaults.items() if v is not None}

        if not envs:
            envs = list(self.envs.keys())

        for env in envs:
            for style in self.style(env):
                if style == "yaml":
                    yield from self._iter_yaml(env, **defaults)
                elif style == "requirements":
                    yield from self._iter_reqs(env, **defaults)
                else:
                    raise ValueError(f"unknown style {style}")  # pragma: no cover

    @classmethod
    def from_toml_dict(
        cls, data: dict[str, Any], user_config: dict[str, Any] | None = None
    ) -> Self:
        """Create from toml dictionaries."""
        data = get_in(["tool", "pyproject2conda"], data, default={})

        if "default_envs" in data:
            default_envs = data.pop("default_envs")

            if "envs" not in data:
                data["envs"] = {}

            for env in default_envs:
                data["envs"][env] = {}

        c = cls(data)

        # add in "default_envs"

        if user_config:  # pragma: no cover
            u = cls.from_toml_dict(data=user_config)
            c = c.assign_user_config(user=u)

        return c

    @classmethod
    def from_string(cls, s: str, user_config: str | None = None) -> Self:
        """Create from string representation of toml file."""
        import tomli

        c = cls.from_toml_dict(tomli.loads(s))

        if user_config:
            u = cls.from_string(user_config)
            c = c.assign_user_config(user=u)
        return c

    @classmethod
    def from_file(
        cls, path: str | Path, user_config: str | Path | None = "infer"
    ) -> Self:
        """Create from toml file(s)."""
        import tomli

        with open(path, "rb") as f:
            data = tomli.load(f)

        c = cls.from_toml_dict(data)

        if user_config == "infer":
            if (user_config := c.user_config()) is not None:
                # relative path
                user_config = Path(path).parent / Path(user_config)

        if user_config:
            if Path(user_config).exists():
                u = cls.from_file(user_config)
                c = c.assign_user_config(u)

        return c



"""Script to create pythons for use with virtualenvs"""
from __future__ import annotations

import sys
from functools import lru_cache

assert sys.version_info >= (3, 9)


@lru_cache
def conda_cmd() -> str:
    import shutil

    if shutil.which("mamba"):
        return "mamba"
    elif shutil.which("conda"):
        return "conda"
    else:
        raise ValueError("must have mamba or conda on path")


def create_env_from_spec(
    env_name: str,
    spec: str | list[str],
    verbose: bool = True,
    flags: str | list[str] | None = None,
) -> None:
    import shlex
    import subprocess

    if not isinstance(spec, str):
        spec = " ".join(spec)

    if flags is not None and not isinstance(flags, str):
        flags = " ".join(flags)

    cmd = f"{conda_cmd()} create -n {env_name} {flags} {spec} "
    if verbose:
        print(cmd)

    out = subprocess.check_call(shlex.split(cmd))
    if out != 0:
        raise RuntimeError(f"failed {cmd}")


def create_environments(
    template: str = "test-{version}",
    versions: str | list[str] | None = None,
    verbose: bool = True,
    flags: str | list[str] | None = None,
    env_map: dict[str, str] | None = None,
) -> None:
    if versions is None:
        versions = ["3.8", "3.9", "3.10", "3.11"]

    if isinstance(versions, str):
        versions = [versions]

    for version in versions:
        env_name = template.format(version=version)

        if env_map and env_name in env_map:
            # environment exists
            if verbose:
                print(
                    f"Skipping environment {env_name}.  Pass `--no-skip` to force recreation."
                )
        else:
            if verbose:
                print(f"Creating environment {env_name}.")

            spec = f"python={version}"
            create_env_from_spec(
                env_name=env_name, spec=spec, flags=flags, verbose=verbose
            )


def main() -> None:
    import argparse

    p = argparse.ArgumentParser(
        description="Create conda environments for creating virtualenvs"
    )

    p.add_argument(
        "-p",
        "--python-version",
        type=str,
        default=["3.8", "3.9", "3.10", "3.11"],
        nargs="+",
        help="""
        python versions to create environments for
        """,
    )
    p.add_argument(
        "-t",
        "--template",
        type=str,
        default="python-{version}",
        help="""
        Template for new environments. The available key is "{version}", which
        will expand to the python version being installed.
        """,
    )
    p.add_argument(
        "--yes", default=False, action="store_true", help="pass `--yes` to conda create"
    )
    p.add_argument("--dry", default=False, action="store_true", help="Do dry run.")
    p.add_argument(
        "-v", dest="verbose", default=False, action="store_true", help="verbose"
    )
    p.add_argument(
        "--skip",
        default=True,
        action=argparse.BooleanOptionalAction,
    )

    args = p.parse_args()

    print(args)

    flags = []
    if args.yes:
        flags.append("--yes")
    if args.dry:
        flags.append("--dry")

    if args.skip:
        from .common_utils import get_conda_environment_map

        env_map = get_conda_environment_map()
    else:
        env_map = None

    create_environments(
        versions=args.python_version,
        template=args.template,
        verbose=args.verbose,
        flags=flags,
        env_map=env_map,
    )


if __name__ == "__main__":
    if __package__ is None:
        # Magic to be able to run script as either
        #   $ python -m tools.create_python
        # or
        #   $ python tools/create_python.py
        from pathlib import Path

        here = Path(__file__).absolute()
        __package__ = here.parent.name

    main()



import sys

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias


if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated


if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


__all__ = [
    "TypeAlias",
    "Annotated",
    "Self",
]



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
import pytest
import pyproject2conda
from pyproject2conda.cli import app
from pyproject2conda.config import Config
from pyproject2conda import utils

from click.testing import CliRunner

# from typer.testing import CliRunner


from pathlib import Path
import tempfile

import filecmp
from textwrap import dedent

ROOT = Path(__file__).resolve().parent / "data"


def do_run(runner, command, *opts, filename=None, must_exist=False):
    if filename is None:
        filename = str(ROOT / "test-pyproject.toml")
    filename = Path(filename)
    if must_exist and not filename.exists():
        raise ValueError(f"filename {filename} does not exist")

    result = runner.invoke(app, [command, "-f", str(filename), *opts])
    return result


def test_template():
    assert utils.filename_from_template(None) is None

    expected = "py38-test.yaml"

    t = utils.filename_from_template(
        "py{py}-{env}", env_name="test", python="3.8", ext="yaml"
    )

    assert t == expected

    t = utils.filename_from_template(
        "py{py}-{env}", env_name="test", python_version="3.8", ext="yaml"
    )

    assert t == expected


def test_option_override():
    toml = """\
    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"


    [project.optional-dependencies]
    test = [
        "pandas",
        "pytest", # p2c: -c conda-forge

    ]

    [tool.pyproject2conda]
    channels = ['conda-forge']
    # these are the same as the default values of `p2c project`
    template_python = "py{py}-{env}"
    template = "hello-{env}"
    style = "yaml"
    # options
    python = ["3.10"]
    # Note that this is relative to the location of pyproject.toml
    user_config = "config/userconfig.toml"
    default_envs = ["test", "dev", "dist-pypi"]

    [tool.pyproject2conda.envs.base]
    extras = []
    style = "yaml"
    python = []

    [tool.pyproject2conda.envs.base2]
    style = "yaml"
    extras = []
    """

    d = Config.from_string(dedent(toml))

    output = list(d.iter(envs=["base"]))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "sort": True,
            "base": True,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "hello-base.yaml",
        },
    )

    output = list(d.iter(envs=["base2"]))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "sort": True,
            "base": True,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "py310-base2.yaml",
            "python": "3.10",
        },
    )

    output = list(d.iter(envs=["base"], template="there-{env}"))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "sort": True,
            "base": True,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": False,
            "remove_whitespace": True,
            "output": "there-base.yaml",
        },
    )

    output = list(d.iter(envs=["base"], allow_empty=True, template="there-{env}"))

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "sort": True,
            "base": True,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": True,
            "remove_whitespace": True,
            "output": "there-base.yaml",
        },
    )

    output = list(
        d.iter(
            envs=["base"],
            allow_empty=True,
            remove_whitespace=False,
            template="there-{env}",
        )
    )

    assert output[0] == (
        "yaml",
        {
            "extras": [],
            "sort": True,
            "base": True,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "reqs": None,
            "deps": None,
            "name": None,
            "channels": ["conda-forge"],
            "allow_empty": True,
            "remove_whitespace": False,
            "output": "there-base.yaml",
        },
    )


def test_dry():
    runner = CliRunner()

    expected = """\
# --------------------
# Creating yaml py310-test-extras.yaml
channels:
  - conda-forge
dependencies:
  - python=3.10
  - conda-forge::pytest
  - pandas
# --------------------
# Creating yaml py311-test-extras.yaml
channels:
  - conda-forge
dependencies:
  - python=3.11
  - conda-forge::pytest
  - pandas
# --------------------
# Creating requirements test-extras.txt
pandas
pytest
    """

    result = do_run(runner, "project", "--dry", "--envs", "test-extras")

    assert result.output == dedent(expected)


def test_config_only_default():
    expected = [
        (
            "yaml",
            {
                "extras": ["test"],
                "sort": True,
                "base": True,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python": "3.8",
                "output": "py38-test.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        )
    ]

    s0 = """
    [tool.pyproject2conda]
    python = ["3.8"]
    default_envs = ["test"]
    """
    s1 = """
    [tool.pyproject2conda]
    python = ["3.8"]

    [tool.pyproject2conda.envs.test]
    extras = true
    """

    s2 = """
    [tool.pyproject2conda.envs.test]
    python = "3.8"


    """

    for s in [s0, s1, s2]:
        c = Config.from_string(s)

        assert list(c.iter()) == expected


def test_config_overrides():
    # test overrides env
    s = """
    [tool.pyproject2conda]
    python = ["3.8"]
    default_envs = ["test"]

    [[tool.pyproject2conda.overrides]]
    envs = ["test"]
    base = false
    """

    c = Config.from_string(s)

    expected = (
        "yaml",
        {
            "extras": ["test"],
            "sort": True,
            "base": False,
            "header": None,
            "overwrite": "check",
            "verbose": None,
            "name": None,
            "channels": None,
            "python": "3.8",
            "output": "py38-test.yaml",
            "deps": None,
            "reqs": None,
            "allow_empty": False,
            "remove_whitespace": True,
        },
    )

    assert list(c.iter())[0] == expected


def test_conifg_overrides_no_envs():
    # test overrides env
    s = """
    [tool.pyproject2conda]
    python = ["3.8"]
    default_envs = ["test"]

    [[tool.pyproject2conda.overrides]]
    base = false
    """

    c = Config.from_string(s)

    with pytest.raises(ValueError):
        c.overrides


def test_config_python_include_version():
    s = """
    [tool.pyproject2conda.envs.test-1]
    extras = ["test"]
    output = "py38-test.yaml"
    python_include = "3.8"
    python_version = "3.8"

    [tool.pyproject2conda.envs."py38-test"]
    extras = ["test"]
    python_include = "3.8"
    python_version = "3.8"
    """

    c = Config.from_string(s)

    expected = [
        (
            "yaml",
            {
                "extras": ["test"],
                "sort": True,
                "base": True,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python_include": "3.8",
                "python_version": "3.8",
                "output": "py38-test.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
        (
            "yaml",
            {
                "extras": ["test"],
                "sort": True,
                "base": True,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python_include": "3.8",
                "python_version": "3.8",
                "output": "py38-test.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
    ]

    assert list(c.iter()) == expected

    # no output:


def test_config_user_config():
    # test overrides env
    s = """
    [tool.pyproject2conda.envs.test]
    python = "3.8"


    """

    s_user = """
    [tool.pyproject2conda.envs.user]
    extras = ["a", "b"]
    python = "3.9"

    [[tool.pyproject2conda.overrides]]
    envs = ["test"]
    base = false
    """

    c = Config.from_string(s, s_user)

    expected = [
        (
            "yaml",
            {
                "extras": ["test"],
                "sort": True,
                "base": False,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python": "3.8",
                "output": "py38-test.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
        (
            "yaml",
            {
                "extras": ["a", "b"],
                "sort": True,
                "base": True,
                "header": None,
                "overwrite": "check",
                "verbose": None,
                "name": None,
                "channels": None,
                "python": "3.9",
                "output": "py39-user.yaml",
                "deps": None,
                "reqs": None,
                "allow_empty": False,
                "remove_whitespace": True,
            },
        ),
    ]

    assert list(c.iter()) == expected

    # blank config, only user
    s = """
    [tool.pyproject2conda]
    """

    c = Config.from_string(s, s_user)

    assert c.data == {
        "envs": {"user": {"extras": ["a", "b"], "python": "3.9"}},
        "overrides": [{"envs": ["test"], "base": False}],
    }


def test_version():
    runner = CliRunner()

    result = runner.invoke(app, ["--version"])

    assert (
        result.stdout.strip()
        == f"pyproject2conda, version {pyproject2conda.__version__}"
    )


def test_multiple():
    runner = CliRunner()

    t1 = tempfile.TemporaryDirectory()
    path1 = t1.name
    # path1 = ROOT / ".." / ".." / "tmp" / "output1"

    do_run(
        runner,
        "project",
        "--template-python",
        f"{path1}/" + "py{py}-{env}",
        "--template",
        f"{path1}/" + "{env}",
    )

    # running this again?
    result = do_run(
        runner,
        "project",
        "-v",
        "--template-python",
        f"{path1}/" + "py{py}-{env}",
        "--template",
        f"{path1}/" + "{env}",
    )

    t2 = tempfile.TemporaryDirectory()
    path2 = t2.name
    # path2 = ROOT / ".." / ".." / "tmp" / "output2"

    do_run(
        runner, "yaml", "-e", "dev", "-p", "3.10", "-o", f"{path2}/py310-dev.yaml", "-v"
    )

    do_run(
        runner,
        "yaml",
        "-e",
        "dist-pypi",
        "--no-base",
        "-p",
        "3.10",
        "-o",
        f"{path2}/py310-dist-pypi.yaml",
    )

    do_run(runner, "yaml", "-e", "test", "-p", "3.10", "-o", f"{path2}/py310-test.yaml")
    do_run(runner, "yaml", "-e", "test", "-p", "3.11", "-o", f"{path2}/py311-test.yaml")

    do_run(
        runner,
        "yaml",
        "-e",
        "test",
        "--no-base",
        "-p",
        "3.10",
        "-o",
        f"{path2}/py310-test-extras.yaml",
    )
    do_run(
        runner,
        "yaml",
        "-e",
        "test",
        "--no-base",
        "-p",
        "3.11",
        "-o",
        f"{path2}/py311-test-extras.yaml",
    )

    do_run(runner, "r", "-e", "test", "--no-base", "-o", f"{path2}/test-extras.txt")

    do_run(
        runner,
        "yaml",
        "-e",
        "dev",
        "-e",
        "dist-pypi",
        "--name",
        "hello",
        "-p",
        "3.10",
        "-d",
        "extra-dep",
        "-r",
        "extra-req",
        "-o",
        f"{path2}/py310-user-dev.yaml",
    )

    do_run(runner, "req", "-o", f"{path2}/base.txt")

    paths1 = Path(path1).glob("*")
    names1 = set(x.name for x in paths1)

    expected = set(
        "base.txt py310-dev.yaml py310-dist-pypi.yaml py310-test-extras.yaml py310-test.yaml py310-user-dev.yaml py311-test-extras.yaml py311-test.yaml test-extras.txt".split()
    )

    assert names1 == expected

    paths2 = Path(path2).glob("*")
    names2 = set(x.name for x in paths2)

    assert expected == names2

    for x in expected:
        assert filecmp.cmp(f"{path1}/{x}", f"{path2}/{x}")

    t1.cleanup()
    t2.cleanup()



"""Utilities to work with nox."""

from __future__ import annotations

import shlex
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Literal, Sequence, TextIO, cast

from ruamel.yaml import YAML

if TYPE_CHECKING:
    from collections.abc import Collection

    import nox


# * Top level installation functions ---------------------------------------------------
def py_prefix(python_version: Any) -> str:
    if isinstance(python_version, str):
        return "py" + python_version.replace(".", "")
    else:
        raise ValueError(f"passed non-string value {python_version}")


def session_environment_filename(
    name: str | None,
    ext: str | None = None,
    python_version: str | None = None,
    lock: bool = False,
) -> str:
    """Get filename for a conda yaml or pip requirements file."""
    if name is None:
        raise ValueError("must supply name")

    filename = name
    if ext is not None:
        filename = filename + ext
    if python_version is not None:
        filename = f"{py_prefix(python_version)}-{filename}"

    if lock:
        if filename.endswith(".yaml"):
            filename = filename.rstrip(".yaml") + "-conda-lock.yml"
        elif filename.endswith(".yml"):
            filename = filename.rstrip(".yml") + "-conda-lock.yml"
        elif filename.endswith(".txt"):
            pass
        else:
            raise ValueError(f"unknown file extension for {filename}")

        return f"./requirements/lock/{filename}"
    else:
        return f"./requirements/{filename}"


def pkg_install_condaenv(
    session: nox.Session,
    name: str,
    lock: bool = False,
    display_name: str | None = None,
    install_package: bool = True,
    update: bool = False,
    update_package: bool = False,
    log_session: bool = False,
    deps: Collection[str] | None = None,
    reqs: Collection[str] | None = None,
    channels: Collection[str] | None = None,
    filename: str | None = None,
    **kwargs: Any,
) -> None:
    """Install requirements.  If need fine control, do it in calling func."""

    def check_filename(filename: str | Path) -> str:
        if not Path(filename).exists():
            raise ValueError(f"file {filename} does not exist")
        session.log(f"Environment file: {filename}")
        return str(filename)

    assert isinstance(session.python, str)
    filename = filename or session_environment_filename(
        name=name, ext=".yaml", python_version=session.python, lock=lock
    )

    if lock:
        session_install_envs_lock(
            session=session,
            lockfile=check_filename(filename),
            display_name=display_name,
            update=update,
            update_package=update_package,
            install_package=install_package,
            **kwargs,
        )

    else:
        session_install_envs(
            session,
            check_filename(filename),
            display_name=display_name,
            update=update,
            update_package=update_package,
            deps=deps,
            reqs=reqs,
            channels=channels,
            install_package=install_package,
            **kwargs,
        )

    if log_session:
        session_log_session(session, conda=True)


def pkg_install_venv(
    session: nox.Session,
    name: str,  # pyright: ignore
    lock: bool = False,
    requirement_paths: Collection[str] | None = None,
    constraint_paths: Collection[str] | None = None,
    extras: str | Collection[str] | None = None,
    reqs: Collection[str] | None = None,
    display_name: str | None = None,
    update: bool = False,
    update_package: bool = False,
    install_package: bool = False,
    no_deps: bool = True,
    log_session: bool = False,
) -> None:
    if lock:
        raise ValueError("lock not yet supported for install_pip")

    session_install_pip(
        session=session,
        requirement_paths=requirement_paths,
        constraint_paths=constraint_paths,
        extras=extras,
        reqs=reqs,
        display_name=display_name,
        update=update,
        update_package=update_package,
        install_package=install_package,
        no_deps=no_deps,
        lock=lock,
    )

    if log_session:
        session_log_session(session, conda=False)


def session_log_session(session: nox.Session, conda: bool = True) -> None:
    logfile = Path(session.create_tmp()) / "env_info.txt"

    session.log(f"writing environment log to {logfile}")

    with logfile.open("w") as f:
        if conda:
            session.run("conda", "list", stdout=f)
        else:
            session.run("python", "--version", stdout=f)
            session.run("pip", "list", stdout=f)


# * User config ------------------------------------------------------------------------
def load_nox_config(path: str | Path = "./config/userconfig.toml") -> dict[str, Any]:
    """
    Load user toml config file.

    File should look something like:

    [nox.python]
    paths = ["~/.conda/envs/python-3.*/bin"]

    # Extras for environments
    # for example, could have
    # dev = ["dev", "nox", "tools"]
    [nox.extras]
    dev = ["dev", "nox"]
    """

    from .projectconfig import ProjectConfig

    return ProjectConfig.from_path(path).to_nox_config()


# * Basic utilities --------------------------------------------------------------------
def combine_list_str(opts: list[str]) -> list[str]:
    if opts:
        return shlex.split(" ".join(opts))
    else:
        return []


def combine_list_list_str(opts: list[list[str]]) -> Iterable[list[str]]:
    return (combine_list_str(opt) for opt in opts)


def sort_like(values: Collection[Any], like: Sequence[Any]) -> list[Any]:
    """Sort `values` in order of `like`."""
    # only unique
    sorter = {k: i for i, k in enumerate(like)}
    return sorted(set(values), key=lambda k: sorter[k])


def update_target(
    target: str | Path, *deps: str | Path, allow_missing: bool = False
) -> bool:
    """Check if target is older than deps:"""
    target = Path(target)

    deps_filtered = []
    for d in map(Path, deps):
        if d.exists():
            deps_filtered.append(d)
        elif not allow_missing:
            raise ValueError(f"dependency {d} does not exist")

    if not target.exists():
        return True
    else:
        target_time = target.stat().st_mtime

        update = any(target_time < dep.stat().st_mtime for dep in deps_filtered)

    return update


def prepend_flag(flag: str, *args: str | Sequence[str]) -> list[str]:
    """
    Add in a flag before each arg.

    >>> prepent_flag("-k", "a", "b")
    ["-k", "a", "-k", "b"]
    """

    args_ = []
    for x in args:
        if isinstance(x, str):
            args_.append(x)
        else:
            args_.extend(x)

    return sum([[flag, _] for _ in args_], [])


def open_webpage(path: str | Path | None = None, url: str | None = None) -> None:
    """
    Open webpage from path or url.

    Useful if want to view webpage with javascript, etc., as well as static html.
    """
    import webbrowser
    from urllib.request import pathname2url

    if path:
        url = "file://" + pathname2url(str(Path(path).absolute()))
    if url:
        webbrowser.open(url)


# * Package install --------------------------------------------------------------------
# ** Utilities
def session_skip_install(session: nox.Session) -> bool:
    """
    Utility to check if we're skipping install and reusing existing venv
    This is a hack and may need to change if upstream changes.
    """
    return session._runner.global_config.no_install and session._runner.venv._reused  # type: ignore


def session_run_commands(
    session: nox.Session, commands: list[list[str]], external: bool = True, **kws: Any
) -> None:
    """Run commands command."""

    if commands:
        kws.update(external=external)
        for opt in combine_list_list_str(commands):
            session.run(*opt, **kws)


def session_set_ipykernel_display_name(
    session: nox.Session, display_name: str | None, check_skip_install: bool = True
) -> None:
    """Rename ipython kernel display name."""
    if not display_name or (check_skip_install and session_skip_install(session)):
        return
    else:
        command = f"python -m ipykernel install --sys-prefix --display-name {display_name}".split()
        # continue if fails
        session.run(*command, success_codes=[0, 1])


def session_install_package(
    session: nox.Session,
    package: str = ".",
    develop: bool = True,
    no_deps: bool = True,
    *args: str,
    **kwargs: Any,
) -> None:
    """Install package into session."""

    if session_skip_install(session):
        return

    if develop:
        command = ["-e"]
    else:
        command = []

    command.append(package)

    if no_deps:
        command.append("--no-deps")

    session.install(*command, *args, **kwargs)


# ** conda-lock
def session_install_envs_lock(
    session: nox.Session,
    lockfile: str | Path,
    extras: str | list[str] | None = None,
    display_name: str | None = None,
    update: bool = False,
    update_package: bool = False,
    install_package: bool = False,
) -> bool:
    """Install dependencies using conda-lock."""

    if session_skip_install(session):
        return True

    unchanged, hashes = env_unchanged(
        session, lockfile, prefix="lock", other=dict(install_package=install_package)
    )

    do_dep = update or (not unchanged)
    do_pkg = install_package and (do_dep or update_package)

    if do_dep:
        if extras:
            if isinstance(extras, str):
                extras = extras.split(",")
            extras = cast(list[str], sum([["--extras", _] for _ in extras], []))
        else:
            extras = []

        session.run(
            "conda-lock",
            "install",
            "--mamba",
            *extras,
            "-p",
            str(session.virtualenv.location),
            str(lockfile),
            silent=True,
            external=True,
        )

    if do_pkg:
        session_install_package(session)

    if do_dep or do_pkg:
        session_set_ipykernel_display_name(session, display_name)
        write_hashfile(hashes, session=session, prefix="lock")

    return unchanged


# ** Conda
def parse_envs(
    *paths: str | Path,
    remove_python: bool = True,
    deps: Collection[str] | None = None,
    reqs: Collection[str] | None = None,
    channels: Collection[str] | None = None,
) -> tuple[set[str], set[str], set[str], str | None]:
    """Parse an `environment.yaml` file."""
    import re

    def _default(x: str | Iterable[str] | None) -> set[str]:
        if x is None:
            return set()
        elif isinstance(x, str):
            x = [x]
        return set(x)

    channels = _default(channels)
    deps = _default(deps)
    reqs = _default(reqs)
    name = None

    python_match = re.compile(r"\s*(python)\s*[~<=>].*")

    def _get_context(path: str | Path | TextIO) -> TextIO | Path:
        if hasattr(path, "readline"):
            from contextlib import nullcontext

            return nullcontext(path)  # type: ignore
        else:
            return Path(path).open("r")

    for path in paths:
        with _get_context(path) as f:
            data = YAML(typ="safe", pure=True).load(f)

        channels.update(data.get("channels", []))
        name = data.get("name", name)

        # check dependencies for pip
        for d in data.get("dependencies", []):
            if isinstance(d, dict):
                reqs.update(cast(list[str], d.get("pip")))
            else:
                if remove_python and not python_match.match(d):
                    deps.add(d)

    return channels, deps, reqs, name


def session_install_envs(
    session: nox.Session,
    *paths: str | Path,
    remove_python: bool = True,
    deps: Collection[str] | None = None,
    reqs: Collection[str] | None = None,
    channels: Collection[str] | None = None,
    conda_install_kws: dict[str, Any] | None = None,
    install_kws: dict[str, Any] | None = None,
    display_name: str | None = None,
    update: bool = False,
    update_package: bool = False,
    install_package: bool = False,
) -> bool:
    """Parse and install everything. Pass an already merged yaml file."""

    if session_skip_install(session):
        return True

    channels, deps, reqs, name = parse_envs(
        *paths,
        remove_python=remove_python,
        deps=deps,
        reqs=reqs,
        channels=channels,
    )

    unchanged, hashes = env_unchanged(
        session,
        prefix="env",
        other=dict(
            deps=deps,
            reqs=reqs,
            channels=channels,
            install_package=install_package,
        ),
    )

    do_dep = update or (not unchanged)
    do_pkg = install_package and (do_dep or update_package)

    if do_dep:
        if not channels:
            channels = ""
        if deps:
            conda_install_kws = conda_install_kws or {}
            conda_install_kws.update(channel=channels)
            if update:
                deps = ["--update-all"] + list(deps)

            session.conda_install(*deps, **(conda_install_kws or {}))

        if reqs:
            if update:
                reqs = ["--upgrade"] + list(reqs)
            session.install(*reqs, **(install_kws or {}))

    if do_pkg:
        session_install_package(session)

    if do_dep or do_pkg:
        session_set_ipykernel_display_name(session, display_name)
        write_hashfile(hashes, session=session, prefix="env")

    return unchanged


# ** Pip
def session_install_pip(
    session: nox.Session,
    requirement_paths: str | Collection[str] | None = None,
    constraint_paths: str | Collection[str] | None = None,
    extras: str | Collection[str] | None = None,
    reqs: str | Collection[str] | None = None,
    display_name: str | None = None,
    update: bool = False,
    update_package: bool = False,
    install_package: bool = False,
    no_deps: bool = True,
    lock: bool = False,
) -> bool:
    if session_skip_install(session):
        return True

    def _check_param(x: None | str | list[str] | Iterable[str]) -> list[str]:
        if x is None:
            return []
        elif isinstance(x, str):
            return [x]
        elif isinstance(x, list):
            return x
        else:
            return list(x)

    def _verify_paths(paths: str | list[str]) -> list[str]:
        if isinstance(paths, str):
            paths = [paths]

        out = []
        for path in paths:
            if Path(path).exists():
                out.append(path)
            else:
                inferred = session_environment_filename(name=path, lock=lock)
                if Path(inferred).exists():
                    out.append(inferred)
                else:
                    raise ValueError(f"no file {path} found/inferred")
        return out

    # parameters
    extras = _check_param(extras)
    if extras:
        install_package = True
        extras = ",".join(extras)
        install_package_args = ["-e", f".[{extras}]"]
    elif install_package:
        install_package_args = ["-e", "."]
    else:
        install_package_args = []

    if install_package and no_deps:
        install_package_args.append("--no-deps")

    requirement_paths = _verify_paths(_check_param(requirement_paths))
    constraint_paths = _verify_paths(_check_param(constraint_paths))
    reqs = _check_param(reqs)
    paths = list(requirement_paths) + list(constraint_paths)

    # check update
    unchanged, hashes = env_unchanged(
        session,
        *paths,
        prefix="pip",
        other=dict(
            reqs=reqs, extras=extras, install_package=install_package, no_deps=no_deps
        ),
    )

    do_dep = update or (not unchanged)
    do_pkg = install_package and (do_dep or update_package)

    if do_dep:
        # do install
        install_args = (
            prepend_flag("-r", *requirement_paths)
            + prepend_flag("-c", *constraint_paths)
            + list(reqs)
        )

        if install_args:
            if update:
                install_args = ["--upgrade"] + list(install_args)
            session.install(*install_args)

    if do_pkg:
        session.install(*install_package_args)

    if do_dep or do_pkg:
        session_set_ipykernel_display_name(session, display_name)
        write_hashfile(hashes, session=session, prefix="pip")

    return unchanged


# ** Hash environment

PREFIX_HASH_EXTS = Literal["env", "lock", "pip"]


def env_unchanged(
    session: nox.Session,
    *paths: str | Path,
    prefix: PREFIX_HASH_EXTS,
    verbose: bool = True,
    hashes: dict[str, str] | None = None,
    other: dict[str, Any] | None = None,
) -> tuple[bool, dict[str, str]]:
    hashfile = hashfile_path(session, prefix)

    if hashes is None:
        hashes = get_hashes(*paths, other=other)

    if hashfile.exists():
        if verbose:
            session.log(f"hash file {hashfile} exists")
        unchanged = hashes == read_hashfile(hashfile)
    else:
        unchanged = False

    if unchanged:
        session.log(f"session {session.name} unchanged")
    else:
        session.log(f"session {session.name} changed")

    return unchanged, hashes


def get_hashes(
    *paths: str | Path,
    other: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Get md5 hashes for paths."""

    out: dict[str, Any] = {"path": {str(path): _get_file_hash(path) for path in paths}}

    if other:
        import hashlib

        other_hashes = {}
        for k, v in other.items():
            if isinstance(v, str):
                s = v
            else:
                try:
                    s = str(sorted(v))
                except Exception:
                    s = str(v)
            other_hashes[k] = hashlib.md5(s.encode("utf-8")).hexdigest()

        out["other"] = other_hashes

    return out


def hashfile_path(session: nox.Session, prefix: PREFIX_HASH_EXTS) -> Path:
    """Path for hashfile for this session."""
    return Path(session.create_tmp()) / f"{prefix}.json"


def write_hashfile(
    hashes: dict[str, str],
    session: nox.Session,
    prefix: PREFIX_HASH_EXTS,
) -> None:
    import json

    path = hashfile_path(session, prefix)

    with open(path, "w") as f:
        json.dump(hashes, f)


def read_hashfile(
    path: str | Path,
) -> dict[str, str]:
    import json

    with open(path) as f:
        data = json.load(f)
    return cast(dict[str, str], data)


def _get_file_hash(path: str | Path, buff_size: int = 65536) -> str:
    import hashlib

    md5 = hashlib.md5()
    with open(path, "rb") as f:
        while True:
            data = f.read(buff_size)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


# * Old stuff --------------------------------------------------------------------------
# def session_install_envs_merge(
#     session,
#     *paths,
#     remove_python=True,
#     deps=None,
#     reqs=None,
#     channels=None,
#     conda_install_kws=None,
#     install_kws=None,
#     display_name=None,
#     update=False,
# ) -> bool:
#     """Merge files (using conda-merge) and then create env"""

#     if session_skip_install(session):
#         return True

#     unchanged, hashes = env_unchanged(
#         session, *paths, prefix="env", other=dict(deps=deps, reqs=reqs)
#     )
#     if unchanged and not update:
#         return unchanged

#     # first create a temporary file for the environment
#     with tempfile.TemporaryDirectory() as d:
#         yaml = Path(d) / "tmp_env.yaml"
#         with yaml.open("w") as f:
#             session.run("conda-merge", *paths, stdout=f, external=True)
#         session.run("cat", str(yaml), external=True, silent=True)

#         channels, deps, reqs, _ = parse_envs(
#             yaml, remove_python=remove_python, deps=deps, reqs=reqs, channels=channels
#         )

#     if deps:
#         if conda_install_kws is None:
#             conda_install_kws = {}
#         conda_install_kws.update(channel=channels)
#         session.conda_install(*deps, **conda_install_kws)

#     if reqs:
#         if install_kws is None:
#             install_kws = {}
#         session.install(*reqs, **install_kws)

#     session_set_ipykernel_display_name(session, display_name)

#     write_hashfile(hashes, session=session, prefix="env")

#     return unchanged


# def _remove_python_from_yaml(path):
#     from yaml import safe_dump

#     path = Path(path)

#     with path.open("r") as f:
#         data = safe_load(f)

#     from copy import deepcopy

#     out = deepcopy(data)

#     for dep in list(out["dependencies"]):
#         if isinstance(dep, str) and dep[: len("python")] == "python":
#             out["dependencies"].remove(dep)

#     path_out = path.with_suffix(".final.yaml")

#     with path_out.open("w") as f:
#         safe_dump(out, f)

#     return path_out


# def session_install_envs_update(
#     session: nox.Session,
#     conda_backend: str,
#     *paths: str | Path,
#     remove_python: bool = True,
#     deps: Sequence[str] | None = None,
#     reqs: Sequence[str] | None = None,
#     conda_install_kws: Mapping[str, str] | None = None,
#     install_kws: Mapping[str, str] | None = None,
#     display_name: str | None = None,
# ) -> None:
#     """Install multiple 'environment.yaml' files."""

#     if session_skip_install(session):
#         return

#     from shutil import which

#     if not which("conda-merge"):
#         session.conda_install("conda-merge")

#     # pin the python version

#     with tempfile.TemporaryDirectory() as d:
#         yaml = Path(d) / "tmp_env.yaml"
#         with yaml.open("w") as f:
#             session.run("conda-merge", *paths, stdout=f, external=True)

#         if remove_python:
#             yaml = _remove_python_from_yaml(yaml)

#         session.run("cat", str(yaml), external=True, silent=False)

#         session.run(
#             conda_backend,
#             "env",
#             "update",
#             "--prefix",
#             session.virtualenv.location,
#             "--file",
#             str(yaml),
#             silent=True,
#             external=True,
#         )

#     session_set_ipykernel_display_name(session, display_name)


# def pin_python_version(session: nox.Session):
#     path = Path(session.virtualenv.location) / "conda-meta" / "pinned"

#     with path.open("w") as f:
#         session.run(
#             "python",
#             "-c",
#             """import sys; print("python=={v.major}.{v.minor}.{v.micro}".format(v=sys.version_info))""",
#             stdout=f,
#         )

# def session_install_envs_update_pin(
#     session: nox.Session,
#     conda_backend: str,
#     *paths: str | Path,
#         display_name: str | None = None,
#     **kws,
# ) -> None:
#     """Install multiple 'environment.yaml' files."""

#     if session_skip_install(session):
#         return

#     from shutil import which

#     if not which("conda-merge"):
#         session.conda_install("conda-merge")

#     # pin the python version
#     pin_python_version(session)

#     with tempfile.TemporaryDirectory() as d:
#         yaml = Path(d) / "tmp_env.yaml"
#         with yaml.open("w") as f:
#             session.run("conda-merge", *paths, stdout=f, external=True)

#         session.run("cat", str(yaml), external=True, silent=False)

#         session.run(
#             conda_backend,
#             "env",
#             "update",
#             "--prefix",
#             session.virtualenv.location,
#             "--file",
#             str(yaml),
#             silent=True,
#             external=True,
#             **kws,
#         )

#     session_set_ipykernel_display_name(session, display_name)


# def parse_args_for_flag(args, flag, action="value"):
#     """
#     Parse args for flag and pop it off args

#     Parameters
#     ----------
#     args : iterable
#         For example, session.posargs.
#     flag : string
#         For example, `flag='--run-external'
#     action : {'value', 'values', 'store_true', 'store_false'}

#     If flag can take multiple values, they should be separated by commas

#     If multiples, return a tuple, else return a string.
#     """
#     flag = flag.strip()
#     n = len(flag)

#     def process_value(arg):
#         if action == "store_true":
#             value = True
#         elif action == "store_false":
#             value = False
#         else:
#             s = arg.split("=")
#             if len(s) != 2:
#                 raise ValueError(f"must supply {flag}=value")
#             if action == "value":
#                 value = s[-1].strip()
#             else:
#                 value = tuple(_.strip() for _ in s[-1].split(","))

#         return value

#     def check_for_flag(arg):
#         s = arg.strip()
#         if action.startswith("value"):
#             return s[:n] == f"{flag}"
#         else:
#             return s == flag

#     # initial value
#     if action == "store_true":
#         value = False
#     elif action == "store_false":
#         value = True
#     elif action in ["value", "values"]:
#         value = None
#     else:
#         raise ValueError(
#             f"action {action} must be one of [store_true, store_false, value, values]"
#         )

#     out = []
#     for arg in args:
#         if check_for_flag(arg):
#             value = process_value(arg)
#         else:
#             out.append(arg)

#     return value, out


# def parse_args_run_external(args):
#     """Parse (and pop) for --run-external flag"""
#     return parse_args_for_flag(args, flag="--run-external", action="store_true")


# def parse_args_test_version(args):
#     """Parse for flag --test-version=..."""
#     return parse_args_for_flag(args, flag="--test-version", action="value")


# def parse_args_pip_extras(args, default=None, join=True):
#     """Parse for flag '--pip-extras=..."""
#     extras, args = parse_args_for_flag(args, flag="--pip-extras", action="values")

#     if extras:
#         extras = set(extras)

#     if default:
#         if extras is None:
#             extras = set()
#         if isinstance(default, str):
#             default = (default,)
#         for d in default:
#             extras.update(d.split(","))

#     if extras and join:
#         extras = ",".join(extras)

#     return extras, args


# def check_args_with_default(args, default=None):
#     """If no args and have a default, place it in args."""
#     if not args and default:
#         if isinstance(default, str):
#             default = default.split()
#         args = default
#     return args


# def run_with_external_check(
#     session, args=None, default=None, check_run_external=True, **kws
# ):
#     """
#     Use session.run with session.posargs.
#     Perform `seesion.run(*args)`, where `args` comes from posargs.
#     If no posargs, then use default.
#     Also, check for flag '--run-external'.  If present,
#     call `session.run(*args, external=True)`
#     """

#     if args is None:
#         args = session.posargs

#     if check_run_external:
#         external, args = parse_args_run_external(args)
#     else:
#         external = False

#     args = check_args_with_default(args, default=default)

# #     session.log(f"args {args}")
# #     session.log(f"external {external}")
# #     session.run(*args, external=external, **kws)


## This should actually go in the noxfile.  Keeping here
## in case want it again in the future.
# @group.session(python=PYTHON_DEFAULT_VERSION)
# def conda_merge(
#     session: Session,
#     conda_merge_force: bool = False,
#     update: FORCE_REINSTALL_CLI = False,
# ):
#     """Merge environments using conda-merge."""
#     import tempfile
#     session_install_envs(
#         session,
#         reqs=["conda-merge", "ruamel.yaml"],
#         update=update,
#     )

#     env_base = ROOT / "environment.yaml"
#     env_dir = ROOT / "environment"

#     def create_env(*extras, others=None, name=None, base=True):
#         if name is None:
#             name = extras[0]
#         env = env_dir / f"{name}.yaml"

#         deps = []
#         if base:
#             deps.append(str(env_base))
#         for extra in extras:
#             deps.append(str(env_dir / f"{extra}-extras.yaml"))

#         if conda_merge_force or update_target(env, *deps):
#             session.log(f"creating {env}")

#             args = ["conda-merge"] + deps
#             with tempfile.TemporaryDirectory() as d:
#                 tmp_path = Path(d) / "tmp_env.yaml"

#                 with tmp_path.open("w") as f:
#                     session.run(*args, stdout=f)

#                 run_str = dedent(
#                     f"""
#                 from ruamel.yaml import YAML; from pathlib import Path;
#                 pin, pout = Path("{tmp_path}"), Path("{env}")
#                 y = YAML(); y.indent(mapping=2, sequence=4, offset=2)
#                 y.dump(y.load(pin.open("r")), pout.open("w"))
#                 """
#                 )

#                 session.run("python", "-c", run_str, silent=True)

#     for extra in ["test", "docs"]:
#         create_env(extra, base=True)

#     create_env("test", "typing", name="typing", base=True)
#     create_env("dev", "test", "typing", "nox", name="dev", base=True)



"""
Top level API (:mod:`pyproject2conda`)
======================================
"""
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _version

try:
    __version__ = _version("pyproject2conda")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "999"

from .parser import PyProject2Conda

__author__ = """William P. Krekelberg"""
__email__ = "wpk@nist.gov"


__all__ = [
    "__version__",
    "PyProject2Conda",
]



from typing import Any, Callable, TypeVar

from ._typing_compat import TypeAlias

FuncType: TypeAlias = Callable[..., Any]

F = TypeVar("F", bound=FuncType)
R = TypeVar("R")

Dec: TypeAlias = Callable[[F], F]



"""
Methods to work with config file (for personal setup) of nox and other project tools.

Right now, this is only for user specific nox config.  But leaving open it could be used
for other things in the future.
"""
from __future__ import annotations

import configparser
import json
import os
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping

if TYPE_CHECKING:
    from typing_extensions import Self


class ProjectConfig:
    """
    Read/write/update userconfig.toml

    This uses a combination of ini/json to read/write the config file.
    This is to avoid having to import from tomli/tomli_w.

    If/when move to python>3.11, can go back to just using toml.


    Parameters
    ----------
    python_paths : list of str, optional
        Paths to append to "PATH" to find python interpreters.
        Note that this can include wildcards.
    env_extras : mapping from str to list of str, option
    """

    def __init__(
        self,
        python_paths: list[str] | None = None,
        env_extras: Mapping[str, Mapping[str, Any]] | None = None,
        copy: bool = True,
    ) -> None:
        python_paths = python_paths or []
        env_extras = {} if env_extras is None else dict(env_extras)

        if copy:
            python_paths = deepcopy(python_paths)
            env_extras = deepcopy(env_extras)

        self.python_paths = python_paths
        self.env_extras = env_extras

    def new_like(
        self,
        python_paths: list[str] | None = None,
        env_extras: Mapping[str, Mapping[str, Any]] | None = None,
        copy: bool = True,
    ) -> Self:
        return type(self)(
            python_paths=python_paths or self.python_paths,
            env_extras=env_extras or self.env_extras,
            copy=copy,
        )

    @staticmethod
    def _path_to_params(
        path: str | Path,
    ) -> tuple[list[str], dict[str, dict[str, Any]]]:
        path = Path(path)

        python_paths: list[str] = []
        env_extras: dict[str, dict[str, Any]] = {}

        if path.exists():
            config = configparser.ConfigParser()
            config.read(path)

            # nox.python
            if "nox.python" in config and "paths" in config["nox.python"]:
                python_paths = json.loads(config["nox.python"]["paths"])

            for header, table in config.items():
                if "tool.pyproject2conda" in header:
                    env_extras[header] = {k: json.loads(v) for k, v in table.items()}

        return python_paths, env_extras

    @classmethod
    def from_path(cls, path: str | Path = "./config/userconfig.toml") -> Self:
        path = Path(path)

        if path.exists():
            python_paths, env_extras = cls._path_to_params(path)
            return cls(python_paths=python_paths, env_extras=env_extras)
        else:
            return cls()

    @staticmethod
    def _params_to_string(
        python_paths: list[str], env_extras: dict[str, Mapping[str, Any]]
    ) -> str:
        import configparser
        import json
        from io import StringIO
        from textwrap import dedent

        config = configparser.ConfigParser()

        if python_paths:
            config.add_section("nox.python")
            config.set("nox.python", "paths", json.dumps(python_paths))

        if env_extras:
            for header, table in env_extras.items():
                config.add_section(header)
                for k, v in table.items():
                    config.set(header, k, json.dumps(v))

        with StringIO() as f:
            config.write(f)
            s = f.getvalue()

        header = dedent(
            """\
        # This file is for setting user specific config for use
        # with nox and other project tools and applications.
        #
        # THIS FILE SHOULD NOT BE TRACKED BY GIT!!!!
        #
        # Example usage:
        #
        # [nox.python]
        # paths = ["~/.conda/envs/python-3.*/bin"]
        #
        # [tool.pyproject2conda.envs.dev]
        # extras = ["dev-complete"]
        """
        )

        return header + s

    def to_path(self, path: str | Path | None = None) -> str:
        s = self._params_to_string(
            python_paths=self.python_paths, env_extras=self.env_extras
        )

        if path is not None:
            with open(path, "w") as f:
                f.write(s)
        return s

    def __repr__(self) -> str:
        return f"<ProjectConfig(python_paths={self.python_paths}, env_extras={self.env_extras})>"

    def expand_python_paths(self) -> list[str]:
        from glob import glob

        paths = []
        for p in self.python_paths:
            paths.extend(glob(os.path.expanduser(p)))
        return paths

    def add_paths_to_environ(
        self, paths: list[str] | None, prepend: bool = True
    ) -> None:
        if paths is None:
            paths = self.expand_python_paths()
        paths_str = ":".join(map(str, paths))
        if prepend:
            fmt = "{path_new}:{path_old}"
        else:
            fmt = "{path_old}:{path_new}"
        os.environ["PATH"] = fmt.format(path_new=paths_str, path_old=os.environ["PATH"])

    def to_nox_config(
        self, add_paths_to_environ: bool = True, prepend: bool = True
    ) -> dict[str, Any]:
        config: dict[str, Any] = {}

        if self.python_paths:
            config["paths"] = self.expand_python_paths()

            if add_paths_to_environ:
                self.add_paths_to_environ(config["paths"], prepend=prepend)

        if self.env_extras:
            config["environment-extras"] = self.env_extras
        else:
            config["environment-extras"] = {"dev": ["nox", "dev"]}

        return config


def glob_envs_to_paths(globs: list[str]) -> list[str]:
    import fnmatch

    from .common_utils import get_conda_environment_map

    env_map = get_conda_environment_map()

    out = []
    for glob in globs:
        found_envs = fnmatch.filter(env_map.keys(), glob)
        out.extend([f"{env_map[k]}/bin" for k in found_envs])

    return out


def main() -> None:
    import argparse

    p = argparse.ArgumentParser(description="Create the file config/userconfig.toml")

    p.add_argument(
        "-p",
        "--python-paths",
        nargs="+",
        help="""
        Specify paths to add to search path for python interpreters.
        This can include the wildcard '*'.
        """,
    )
    p.add_argument(
        "-e",
        "--env",
        nargs="+",
        help="""
        Conda environment name patterns to extract `--python-paths` from
        """,
    )
    p.add_argument(
        "-d",
        "--dev-extras",
        nargs="+",
        help="extras (from pyproject.toml) to include in development environment",
    )

    p.add_argument(
        "-f",
        "--file",
        help="file to store configuration",
        type=str,
        default="./config/userconfig.toml",
    )

    args = p.parse_args()

    n = ProjectConfig.from_path(path=args.file)

    if args.python_paths:
        python_paths = args.python_paths
    elif args.env:
        python_paths = glob_envs_to_paths(args.env)
    else:
        python_paths = None

    if python_paths:
        n.python_paths = python_paths

    if args.dev_extras:
        n.env_extras["tool.pyproject2conda.envs.dev"] = {"extras": args.dev_extras}

    n.to_path(args.file)


if __name__ == "__main__":
    if __package__ is None:
        # Magic to be able to run script as either
        #   $ python -m tools.create_python
        # or
        #   $ python tools/create_python.py
        here = Path(__file__).absolute()
        __package__ = here.parent.name

    main()



# mypy: disable-error-code="no-untyped-def, no-untyped-call"
from __future__ import annotations
import tempfile
import pytest
from pyproject2conda import parser
from pyproject2conda.utils import get_in

from textwrap import dedent


def test_get_in():
    d = {"a": {"b": {"c": 3}}}

    assert get_in(["a", "b", "c"], d) == 3

    assert get_in(["a", "d"], d, "hello") == "hello"


def test_list_to_string():
    assert parser._list_to_str(["a", "b"], eol=True) == "a\nb\n"
    assert parser._list_to_str(["a", "b"], eol=False) == "a\nb"
    assert parser._list_to_str(None) == ""


def test_match_p2c_comment():
    expected = "-c -d"
    for comment in [
        "#p2c: -c -d",
        "# p2c: -c -d",
        "  #p2c: -c -d",
        "# p2c: -c -d # some other thing",
        "# some other thing # p2c: -c -d # another thing",
    ]:
        match = parser._match_p2c_comment(comment)

        assert match == expected

    # check for skip
    for comment in [
        "##p2c: -c -d",
        "## p2c: -c -d",
        "  ##p2c: -c -d",
        "## p2c: -c -d # some other thing",
        "# some other thing ## p2c: -c -d # another thing",
    ]:
        match = parser._match_p2c_comment(comment)

        assert match is None


def test_parse_p2c():
    def get_expected(pip=False, skip=False, channel=None, package=None):
        if package is None:
            package = []
        return {
            "pip": pip,
            "skip": skip,
            "channel": channel,
            "package": package,
        }

    assert parser._parse_p2c(None) is None

    assert parser._parse_p2c("--pip") == get_expected(pip=True)
    assert parser._parse_p2c("-p") == get_expected(pip=True)

    assert parser._parse_p2c("--skip") == get_expected(skip=True)
    assert parser._parse_p2c("-s") == get_expected(skip=True)

    assert parser._parse_p2c("-s -c conda-forge") == get_expected(
        skip=True, channel="conda-forge"
    )

    assert parser._parse_p2c("athing>=0.3,<0.2 ") == get_expected(
        package=["athing>=0.3,<0.2"]
    )

    assert parser._parse_p2c("athing>=0.3,<0.2 bthing ") == get_expected(
        package=["athing>=0.3,<0.2", "bthing"]
    )

    assert parser._parse_p2c("'athing >= 0.3, <0.2' bthing ") == get_expected(
        package=["athing >= 0.3, <0.2", "bthing"]
    )


def test_value_comment_pairs():
    d: list[tuple[str | None, str | None]] = [(None, "# p2c: --pip")]

    with pytest.raises(ValueError):
        out = parser.value_comment_pairs_to_conda(d)

    d = [("hello", "# p2c: there")]

    out = parser.value_comment_pairs_to_conda(d)

    assert out["dependencies"] == ["hello", "there"]


def test_header():
    expected = dedent(
        """\
#
# This file is autogenerated by pyproject2conda.
# You should not manually edit this file.
# Instead edit the corresponding pyproject.toml file.
#"""
    )

    assert expected == parser._create_header()

    cmd = "hello"
    out = parser._create_header(cmd=cmd)

    header = dedent(
        f"""\
#
# This file is autogenerated by pyproject2conda
# with the following command:
#
#     $ {cmd}
#
# You should not manually edit this file.
# Instead edit the corresponding pyproject.toml file.
#"""
    )

    assert out == header


def test_yaml_to_str():
    d = {"dep": ["a", "b"]}

    s = parser._yaml_to_string(d, add_final_eol=False)

    expected = """\
    dep:
      - a
      - b"""

    assert s == dedent(expected)

    s = parser._yaml_to_string(d, add_final_eol=True)

    expected = """\
    dep:
      - a
      - b
    """

    assert s == dedent(expected)


def test_optional_write():
    from pathlib import Path

    s = "hello"
    with tempfile.TemporaryDirectory() as d:
        p = Path(d) / "tmp.txt"

        with open(p, "w") as f:
            parser._optional_write(s, f)

        with open(p, "r") as f:
            test = f.read()

        assert test == s


def test_output_to_yaml():
    s = parser._output_to_yaml(
        dependencies=["a"],
        channels=["conda-forge"],
        pip=["pip-thing"],
        name="hello",
    )

    expected = """\
name: hello
channels:
  - conda-forge
dependencies:
  - a
  - pip
  - pip:
      - pip-thing
    """

    assert s == dedent(expected)


def test_infer():
    toml = dedent(
        """\
    [project]
    name = "hello"
    dependencies = [
    "athing", # p2c: -p # a comment
    "bthing", # p2c: -s bthing-conda
    "cthing; python_version<'3.10'", # p2c: -c conda-forge
    ]

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest", # p2c: -c conda-forge
    ]
    dev-extras = [
    # p2c: -s additional-thing # this is an additional conda package
    "matplotlib", # p2c: -s conda-matplotlib
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]
    dist-pypi = [
    "setuptools",
    "build", # p2c: -p
    ]


    [tool.pyproject2conda]
    channels = ['conda-forge']
    """
    )

    d = parser.PyProject2Conda.from_string(toml)
    with pytest.raises(ValueError):
        d.to_conda_yaml(python_include="infer")


def test_package_name():
    toml = dedent(
        """\
    [project]
    requires-python = ">=3.8,<3.11"
    dependencies = [
    "athing", # p2c: -p # a comment
    "bthing", # p2c: -s bthing-conda
    "cthing; python_version<'3.10'", # p2c: -c conda-forge
    ]

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest", # p2c: -c conda-forge
    ]
    dev-extras = [
    # p2c: -s additional-thing # this is an additional conda package
    "matplotlib", # p2c: -s conda-matplotlib
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]
    dist-pypi = [
    "setuptools",
    "build", # p2c: -p
    ]


    [tool.pyproject2conda]
    channels = ['conda-forge']
    """
    )
    d = parser.PyProject2Conda.from_string(toml)
    with pytest.raises(ValueError):
        d.to_conda_lists()


def test_complete():
    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8, <3.11"
    dependencies = [
    "athing", # p2c: -p # a comment
    "bthing", # p2c: -s bthing-conda
    "cthing; python_version<'3.10'", # p2c: -c conda-forge
    ]

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest", # p2c: -c conda-forge
    ]
    dev-extras = [
    # p2c: -s additional-thing # this is an additional conda package
    "matplotlib", # p2c: -s conda-matplotlib
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]
    dist-pypi = [
    "setuptools",
    "build", # p2c: -p
    ]


    [tool.pyproject2conda]
    channels = ['conda-forge']
    """
    )

    d = parser.PyProject2Conda.from_string(toml)

    # test unknown extra
    with pytest.raises(ValueError):
        d.to_conda_yaml(extras="a-thing")

    # test list:
    assert d.list_extras() == ["test", "dev-extras", "dev", "dist-pypi"]

    assert d._get_opts("hello", "there") == []

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == d.to_conda_yaml()

    # test -p option
    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8,<3.11
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(python_include="infer")

    # with remove spaces:
    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8, <3.11
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="infer", remove_whitespace=False
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.9
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(python_include="python=3.9")

    # test passing python_version
    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.10
  - bthing-conda
  - pip
  - pip:
      - athing
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="python=3.10", python_version="3.10"
    )

    out = d.to_conda_yaml(channels="hello")

    expected = """\
channels:
  - hello
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == out

    out = d.to_conda_yaml(extras="test", sort=False)

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == out

    out = d.to_conda_yaml(extras="test", sort=True)

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - pandas
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == out

    out = d.to_conda_yaml(extras="dist-pypi", include_base_dependencies=False)

    expected = """\
channels:
  - conda-forge
dependencies:
  - setuptools
  - pip
  - pip:
      - build
    """

    assert out == dedent(expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - additional-thing
  - conda-matplotlib
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == d.to_conda_yaml("dev", sort=False)

    expected = """\
channels:
  - conda-forge
dependencies:
  - additional-thing
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - conda-matplotlib
  - pandas
  - pip
  - pip:
      - athing
    """

    assert dedent(expected) == d.to_conda_yaml("dev")

    # Test deps/reqs
    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - dep
  - pip
  - pip:
      - athing
      - req
    """

    assert dedent(expected) == d.to_conda_yaml(
        deps=["dep;python_version<'3.10'"], reqs=["req"]
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.10
  - bthing-conda
  - pip
  - pip:
      - athing
      - req;python_version<'3.10'
    """
    assert dedent(expected) == d.to_conda_yaml(
        deps=["dep;python_version<'3.10'"],
        reqs=["req;python_version<'3.10'"],
        python_include="python=3.10",
        python_version="3.10",
    )


def test_missing_dependencies():
    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"
    dependencies = []

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest", # p2c: -c conda-forge
    "pytest-accept", # p2c: -p
    ]
    dev-extras = [
    # p2c: -s additional-thing # this is an additional conda package
    "matplotlib", # p2c: -s conda-matplotlib
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]

    [tool.pyproject2conda]
    channels = ['conda-forge']
    """
    )

    d = parser.PyProject2Conda.from_string(toml)

    expected = """\
channels:
  - conda-forge
dependencies:
  - conda-forge::pytest
  - pandas
  - pip
  - pip:
      - pytest-accept
    """

    assert dedent(expected) == d.to_conda_yaml(
        extras="test",
        include_base_dependencies=False,
    )

    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8,<3.11"

    [project.optional-dependencies]
    test = [
    "pandas",
    "pytest", # p2c: -c conda-forge
    "pytest-accept", # p2c: -p
    ]
    dev-extras = [
    # p2c: -s additional-thing # this is an additional conda package
    "matplotlib", # p2c: -s conda-matplotlib
    ]
    dev = [
    "hello[test]",
    "hello[dev-extras]",
    ]

    [tool.pyproject2conda]
    channels = ['conda-forge']
    """
    )

    d = parser.PyProject2Conda.from_string(toml)

    assert dedent(expected) == d.to_conda_yaml(
        extras="test",
        include_base_dependencies=False,
    )

    assert dedent(expected) == d.to_conda_yaml(
        extras="test",
        include_base_dependencies=True,
    )

    # check output has no dependencies:
    for attr in ["to_conda_yaml", "to_requirements"]:
        f = getattr(d, attr)
        with pytest.raises(ValueError):
            f()

        assert f(allow_empty=True) == "No dependencies for this environment\n"


def test_spaces():
    toml = dedent(
        """\
    [project]
    name = "hello"
    requires-python = ">=3.8, <3.11"
    dependencies = [
    "athing >0.5", # p2c: -p # a comment
    "bthing > 1.0", # p2c: -s 'bthing-conda > 2.0'
    "cthing; python_version < '3.10'", # p2c: -c conda-forge
    ]


    [tool.pyproject2conda]
    channels = ['conda-forge']
    """
    )

    d = parser.PyProject2Conda.from_string(toml)

    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8, <3.11
  - bthing-conda > 2.0
  - conda-forge::cthing
  - pip
  - pip:
      - athing >0.5
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="infer", remove_whitespace=False
    )

    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8,<3.11
  - bthing-conda>2.0
  - conda-forge::cthing
  - pip
  - pip:
      - athing>0.5
    """
    assert dedent(expected) == d.to_conda_yaml(
        python_include="infer", remove_whitespace=True
    )

    expected = """\
    athing >0.5
    bthing > 1.0
    cthing; python_version < '3.10'
    """

    assert dedent(expected) == d.to_requirements(remove_whitespace=False)

    expected = """\
    athing>0.5
    bthing>1.0
    cthing;python_version<'3.10'
    """

    assert dedent(expected) == d.to_requirements(remove_whitespace=True)



# pyright: reportUnknownMemberType=false, reportGeneralTypeIssues=false
"""
Parse `pyproject.toml` (:mod:`~pyproject2conda.parser`)
=======================================================

Main parser to turn `pyproject.toml` to other formats.
"""
from __future__ import annotations

import argparse
import re
import shlex
from functools import lru_cache
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    Iterable,
    Optional,
    Sequence,
    TextIO,
    TypeVar,
    Union,
    cast,
)

if TYPE_CHECKING:
    import tomlkit.items
    import tomlkit.toml_document

    from ._typing_compat import Self

import tomlkit
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from ruamel.yaml import YAML

from pyproject2conda.utils import get_in

# -- typing ----------------------------------------------------------------------------

Tstr_opt = Optional[str]
Tstr_seq_opt = Optional[Union[str, Sequence[str]]]

T = TypeVar("T")


# -- Utilities -------------------------------------------------------------------------


def _check_allow_empty(allow_empty: bool) -> str:
    msg = "No dependencies for this environment\n"
    if allow_empty:
        return msg
    else:
        raise ValueError(msg)


_WHITE_SPACE_REGEX = re.compile(r"\s+")


def _remove_whitespace(s: str) -> str:
    return re.sub(_WHITE_SPACE_REGEX, "", s)


# --- Default parser -------------------------------------------------------------------


@lru_cache
def _default_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Parser searches for comments '# p2c: [OPTIONS] CONDA-PACKAGES"
    )

    parser.add_argument(
        "-c",
        "--channel",
        type=str,
        help="Channel to add to the pyproject requirement",
    )
    parser.add_argument(
        "-p",
        "--pip",
        action="store_true",
        help="If specified, install pyproject dependency with pip",
    )
    parser.add_argument(
        "-s",
        "--skip",
        action="store_true",
        help="If specified skip pyproject dependency on this line",
    )

    parser.add_argument("package", nargs="*")

    return parser


def _factory_empty_tomlkit_Array() -> tomlkit.items.Array:
    return tomlkit.items.Array([], tomlkit.items.Trivia())


def _unique_list(values: Iterable[T]) -> list[T]:
    """
    Return only unique values in list.
    Unlike using set(values), this preserves order.
    """
    output: list[T] = []
    for v in values:
        if v not in output:
            output.append(v)
    return output


def _list_to_str(values: Iterable[str] | None, eol: bool = True) -> str:
    if values:
        output = "\n".join(values)
        if eol:
            output += "\n"
    else:
        output = ""

    return output


def _iter_value_comment_pairs(
    array: tomlkit.items.Array,  # pyright: ignore
) -> Generator[tuple[Tstr_opt, Tstr_opt], None, None]:
    """Extract value and comments from array"""
    for v in array._value:  # pyright: ignore
        if v.value is not None and not isinstance(
            v.value, tomlkit.items.Null
        ):  # pyright: ignore
            value = str(v.value)  # pyright: ignore
        else:
            value = None
        if v.comment:  # pyright: ignore
            comment = v.comment.as_string()
        else:
            comment = None
        if value is None and comment is None:
            continue
        yield (value, comment)


def _matches_package_name(
    dep: Tstr_opt,
    package_name: str,
) -> list[str] | None:
    """
    Check if `dep` matches pattern {package_name}[extra,..]

    If it does, return extras, else return None
    """

    if not dep:
        return None

    pattern = rf"{package_name}\[(.*?)\]"
    match = re.match(pattern, dep)

    if match:
        extras = match.group(1).split(",")
    else:
        extras = None
    return extras


def _get_value_comment_pairs(
    package_name: str,
    deps: tomlkit.items.Array,
    extras: Tstr_seq_opt = None,
    opts: tomlkit.items.Table | None = None,
    include_base_dependencies: bool = True,
) -> list[tuple[Tstr_opt, Tstr_opt]]:
    """Recursively build dependency, comment pairs from deps and extras."""
    if include_base_dependencies:
        out = list(_iter_value_comment_pairs(deps))
    else:
        out = []

    if extras is None:
        return out
    else:
        assert opts is not None

    if isinstance(extras, str):
        extras = [extras]

    for extra in extras:
        for value, comment in _iter_value_comment_pairs(
            cast(tomlkit.items.Array, opts[extra])
        ):
            if new_extras := _matches_package_name(value, package_name):
                out.extend(
                    _get_value_comment_pairs(
                        package_name=package_name,
                        extras=new_extras,
                        deps=deps,
                        opts=opts,
                        include_base_dependencies=False,
                    )
                )
            else:
                out.append((value, comment))

    return out


def _pyproject_to_value_comment_pairs(
    data: tomlkit.toml_document.TOMLDocument,
    extras: Tstr_seq_opt = None,
    unique: bool = True,
    include_base_dependencies: bool = True,
) -> list[tuple[Tstr_opt, Tstr_opt]]:
    package_name = cast("str | None", get_in(["project", "name"], data, default=None))

    if package_name is None:
        raise ValueError("Must specify `project.package_name` in pyproject.toml")

    deps = cast(
        tomlkit.items.Array,
        get_in(["project", "dependencies"], data, factory=_factory_empty_tomlkit_Array),
    )

    value_comment_list = _get_value_comment_pairs(
        package_name=package_name,
        extras=extras,
        deps=deps,
        opts=get_in(
            ["project", "optional-dependencies"],
            data,
            factory=_factory_empty_tomlkit_Array,
        ),
        include_base_dependencies=include_base_dependencies,
    )

    if unique:
        value_comment_list = _unique_list(value_comment_list)

    return value_comment_list


def _match_p2c_comment(comment: Tstr_opt) -> Tstr_opt:
    if not comment or not (match := re.match(r".*?#\s*p2c:\s*([^\#]*)", comment)):
        return None
    elif re.match(r".*?##\s*p2c:", comment):
        # This checks for double ##.  If found, ignore line
        return None
    else:
        return match.group(1).strip()


def _parse_p2c(match: Tstr_opt) -> dict[str, Any] | None:
    """Parse match from _match_p2c_comment"""

    if match:
        return vars(_default_parser().parse_args(shlex.split(match)))
    else:
        return None


def parse_p2c_comment(comment: Tstr_opt) -> dict[str, Any] | None:
    if match := _match_p2c_comment(comment):
        return _parse_p2c(match)
    else:
        return None


def _limit_deps_by_python_version(
    deps: list[str], python_version: Tstr_opt = None
) -> list[str]:
    if python_version:
        version = Version(python_version)
    else:
        version = None

    matcher = re.compile(
        r"(?P<dep>.*?);\s*python_version\s*(?P<token>[<=>~]*)\s*[\'|\"](?P<version>.*?)[\'|\"]"
    )

    output: list[str] = []
    for dep in deps:
        if match := matcher.match(dep):
            if not version or version in SpecifierSet(
                match["token"] + match["version"]
            ):
                output.append(match["dep"])
        else:
            output.append(dep)
    return output


def value_comment_pairs_to_conda(
    value_comment_list: list[tuple[Tstr_opt, Tstr_opt]],
    sort: bool = True,
    deps: Sequence[str] | None = None,
    reqs: Sequence[str] | None = None,
) -> dict[str, Any]:
    """Convert raw value/comment pairs to install lines"""

    conda_deps: list[str] = []
    pip_deps: list[str] = []

    def _check_value(value: Any) -> None:
        if not value:
            raise ValueError("trying to add value that does not exist")

    for value, comment in value_comment_list:
        if comment and (parsed := parse_p2c_comment(comment)):
            if parsed["pip"]:
                _check_value(value)
                pip_deps.append(value)  # type: ignore
            elif not parsed["skip"]:
                _check_value(value)

                if parsed["channel"]:
                    conda_deps.append("{}::{}".format(parsed["channel"], value))
                else:
                    conda_deps.append(value)  # type: ignore

            conda_deps.extend(parsed["package"])
        elif value:
            conda_deps.append(value)

    if deps:
        conda_deps.extend(list(deps))

    if reqs:
        pip_deps.extend(list(reqs))

    if sort:
        conda_deps = sorted(conda_deps)
        pip_deps = sorted(pip_deps)

    return {"dependencies": conda_deps, "pip": pip_deps}


def pyproject_to_conda_lists(
    data: tomlkit.toml_document.TOMLDocument,
    extras: Tstr_seq_opt = None,
    channels: Tstr_seq_opt = None,
    python_include: Tstr_opt = None,
    python_version: Tstr_opt = None,
    include_base_dependencies: bool = True,
    sort: bool = True,
    deps: Sequence[str] | None = None,
    reqs: Sequence[str] | None = None,
    remove_whitespace: bool = True,
) -> dict[str, Any]:
    if python_include == "infer":
        # safer get
        if (x := get_in(["project", "requires-python"], data)) is None:
            raise ValueError("No value for `requires-python` in pyproject.toml file")
        else:
            python_include = "python" + x.unwrap()

    if channels is None:
        channels_doc = get_in(["tool", "pyproject2conda", "channels"], data, None)
        if channels_doc:
            channels = channels_doc.unwrap()
    if isinstance(channels, str):
        channels = [channels]

    value_comment_list = _pyproject_to_value_comment_pairs(
        data=data,
        extras=extras,
        include_base_dependencies=include_base_dependencies,
    )

    output = value_comment_pairs_to_conda(
        value_comment_list,
        sort=sort,
        deps=deps,
        reqs=reqs,
    )

    if python_include:
        output["dependencies"].insert(0, python_include)
    if channels:
        output["channels"] = channels

    # limit python version/remove python_verions <=> part
    output["dependencies"] = _limit_deps_by_python_version(
        output["dependencies"], python_version
    )

    if remove_whitespace:
        output = {k: [_remove_whitespace(x) for x in v] for k, v in output.items()}

    return output


def pyproject_to_conda(
    data: tomlkit.toml_document.TOMLDocument,
    extras: Tstr_seq_opt = None,
    channels: Tstr_seq_opt = None,
    name: Tstr_opt = None,
    python_include: Tstr_opt = None,
    stream: str | Path | TextIO | None = None,
    python_version: Tstr_opt = None,
    include_base_dependencies: bool = True,
    header_cmd: Tstr_opt = None,
    sort: bool = True,
    deps: Sequence[str] | None = None,
    reqs: Sequence[str] | None = None,
    allow_empty: bool = False,
    remove_whitespace: bool = True,
) -> str:
    output = pyproject_to_conda_lists(
        data=data,
        extras=extras,
        channels=channels,
        python_include=python_include,
        python_version=python_version,
        include_base_dependencies=include_base_dependencies,
        sort=sort,
        deps=deps,
        reqs=reqs,
        remove_whitespace=remove_whitespace,
    )
    return _output_to_yaml(
        **output,
        name=name,
        stream=stream,
        header_cmd=header_cmd,
        allow_empty=allow_empty,
    )


def _create_header(cmd: str | None = None) -> str:
    from textwrap import dedent

    if cmd:
        header = dedent(
            f"""
        This file is autogenerated by pyproject2conda
        with the following command:

            $ {cmd}

        You should not manually edit this file.
        Instead edit the corresponding pyproject.toml file.
        """
        )
    else:
        header = dedent(
            """
        This file is autogenerated by pyproject2conda.
        You should not manually edit this file.
        Instead edit the corresponding pyproject.toml file.
        """
        )

    # prepend '# '
    lines = []
    for line in header.split("\n"):
        if len(line.strip()) == 0:
            lines.append("#")
        else:
            lines.append("# " + line)
    header = "\n".join(lines)
    # header = "\n".join(["# " + line for line in header.split("\n")])
    return header


def _add_header(string: str, header_cmd: str | None) -> str:
    if header_cmd is not None:
        return _create_header(header_cmd) + "\n" + string
    else:
        return string


def _yaml_to_string(
    data: dict[str, Any],
    yaml: Any = None,
    add_final_eol: bool = False,
    header_cmd: str | None = None,
) -> str:
    import io

    if yaml is None:
        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)

    buf = io.BytesIO()
    yaml.dump(data, buf)

    val = buf.getvalue()

    if not add_final_eol:
        val = val[:-1]
    return _add_header(val.decode("utf-8"), header_cmd)


def _optional_write(
    string: str, stream: str | Path | TextIO | None, mode: str = "w"
) -> None:
    if stream is None:
        return
    if isinstance(stream, (str, Path)):
        with open(stream, mode) as f:
            f.write(string)
    else:
        stream.write(string)


def _output_to_yaml(
    dependencies: list[str] | None,
    channels: list[str] | None = None,
    pip: list[str] | None = None,
    name: Tstr_opt = None,
    stream: str | Path | TextIO | None = None,
    header_cmd: str | None = None,
    allow_empty: bool = False,
) -> str:
    data: dict[str, Any] = {}
    if name:
        data["name"] = name

    if channels:
        data["channels"] = channels

    data["dependencies"] = []
    if dependencies:
        data["dependencies"].extend(dependencies)
    if pip:
        data["dependencies"].append("pip")
        data["dependencies"].append({"pip": pip})

    if not data["dependencies"]:
        return _check_allow_empty(allow_empty)
    else:
        # return data
        s = _yaml_to_string(data, add_final_eol=True, header_cmd=header_cmd)
        _optional_write(s, stream)
        return s


class PyProject2Conda:
    """Wrapper class to transform pyproject.toml -> environment.yaml"""

    def __init__(
        self,
        data: tomlkit.toml_document.TOMLDocument,
        name: Tstr_opt = None,
        channels: Tstr_seq_opt = None,
        python_include: Tstr_opt = None,
    ) -> None:
        self.data = data
        self.name = name
        self.channels = channels
        self.python_include = python_include

    def to_conda_yaml(
        self,
        extras: Tstr_seq_opt = None,
        name: Tstr_opt = None,
        channels: Tstr_seq_opt = None,
        python_include: Tstr_opt = None,
        stream: str | Path | TextIO | None = None,
        python_version: Tstr_opt = None,
        include_base_dependencies: bool = True,
        header_cmd: str | None = None,
        sort: bool = True,
        deps: Sequence[str] | None = None,
        reqs: Sequence[str] | None = None,
        allow_empty: bool = False,
        remove_whitespace: bool = True,
    ) -> str:
        self._check_extras(extras)

        return pyproject_to_conda(
            data=self.data,
            extras=extras,
            name=name or self.name,
            channels=channels or self.channels,
            python_include=python_include or self.python_include,
            stream=stream,
            python_version=python_version,
            include_base_dependencies=include_base_dependencies,
            header_cmd=header_cmd,
            sort=sort,
            deps=deps,
            reqs=reqs,
            allow_empty=allow_empty,
            remove_whitespace=remove_whitespace,
        )

    def to_conda_lists(
        self,
        extras: Tstr_seq_opt = None,
        channels: Tstr_seq_opt = None,
        python_include: Tstr_opt = None,
        python_version: Tstr_opt = None,
        include_base_dependencies: bool = True,
        sort: bool = True,
        deps: Sequence[str] | None = None,
        reqs: Sequence[str] | None = None,
        remove_whitespace: bool = True,
    ) -> dict[str, Any]:
        self._check_extras(extras)

        return pyproject_to_conda_lists(
            data=self.data,
            extras=extras,
            channels=channels or self.channels,
            python_include=python_include or self.python_include,
            python_version=python_version,
            include_base_dependencies=include_base_dependencies,
            sort=sort,
            deps=deps,
            reqs=reqs,
            remove_whitespace=remove_whitespace,
        )

    def to_requirement_list(
        self,
        extras: Tstr_seq_opt = None,
        include_base_dependencies: bool = True,
        sort: bool = True,
        reqs: Sequence[str] | None = None,
        remove_whitespace: bool = True,
    ) -> list[str]:
        self._check_extras(extras)

        values = _pyproject_to_value_comment_pairs(
            data=self.data,
            extras=extras,
            include_base_dependencies=include_base_dependencies,
        )
        out = [x for x, _ in values if x is not None]

        if reqs:
            out.extend(list(reqs))

        if remove_whitespace:
            out = [_remove_whitespace(x) for x in out]

        if sort:
            return sorted(out)
        else:
            return out

    def to_requirements(
        self,
        extras: Tstr_seq_opt = None,
        include_base_dependencies: bool = True,
        header_cmd: str | None = None,
        stream: str | Path | TextIO | None = None,
        sort: bool = True,
        reqs: Sequence[str] | None = None,
        allow_empty: bool = False,
        remove_whitespace: bool = True,
    ) -> str:
        """Create requirements.txt like file with pip dependencies."""

        self._check_extras(extras)

        reqs = self.to_requirement_list(
            extras=extras,
            include_base_dependencies=include_base_dependencies,
            sort=sort,
            reqs=reqs,
            remove_whitespace=remove_whitespace,
        )

        if not reqs:
            return _check_allow_empty(allow_empty)
        else:
            s = _add_header(_list_to_str(reqs), header_cmd)
            _optional_write(s, stream)
            return s

    def to_conda_requirements(
        self,
        extras: Tstr_seq_opt = None,
        channels: Tstr_seq_opt = None,
        python_include: Tstr_opt = None,
        python_version: Tstr_opt = None,
        prepend_channel: bool = False,
        stream_conda: str | Path | TextIO | None = None,
        stream_pip: str | Path | TextIO | None = None,
        include_base_dependencies: bool = True,
        header_cmd: Tstr_opt = None,
        sort: bool = True,
        deps: Sequence[str] | None = None,
        reqs: Sequence[str] | None = None,
        remove_whitespace: bool = True,
    ) -> tuple[str, str]:
        output = self.to_conda_lists(
            extras=extras,
            channels=channels,
            python_include=python_include,
            python_version=python_version,
            include_base_dependencies=include_base_dependencies,
            sort=sort,
            deps=deps,
            reqs=reqs,
            remove_whitespace=remove_whitespace,
        )

        deps = output.get("dependencies", None)
        reqs = output.get("pip", None)

        channels = output.get("channels", None)
        if channels and prepend_channel:
            assert len(channels) == 1
            channel = channels[0]
            # add in channel if none exists
            if deps:
                deps = [dep if "::" in dep else f"{channel}::{dep}" for dep in deps]

        deps_str = _add_header(_list_to_str(deps), header_cmd)
        reqs_str = _add_header(_list_to_str(reqs), header_cmd)

        if stream_conda and deps_str:
            _optional_write(deps_str, stream_conda)

        if stream_pip and reqs_str:
            _optional_write(reqs_str, stream_pip)

        return deps_str, reqs_str

    def _check_extras(self, extras: Tstr_seq_opt) -> None:
        if extras is None:
            return
        elif isinstance(extras, str):
            sent: Sequence[str] = [extras]
        else:
            sent = extras

        available = self.list_extras()

        for s in sent:
            if s not in available:
                raise ValueError(f"{s} not in {available}")

    def _get_opts(self, *keys: str) -> list[str]:
        opts = get_in(keys, self.data, None)
        if opts:
            return list(opts.keys())
        else:
            return []

    def list_extras(self) -> list[str]:
        return self._get_opts("project", "optional-dependencies")

    @classmethod
    def from_string(
        cls,
        toml_string: str,
        name: Tstr_opt = None,
        channels: Tstr_seq_opt = None,
        python_include: Tstr_opt = None,
    ) -> Self:
        data = tomlkit.parse(toml_string)
        return cls(
            data=data, name=name, channels=channels, python_include=python_include
        )

    @classmethod
    def from_path(
        cls,
        path: str | Path,
        name: Tstr_opt = None,
        channels: Tstr_seq_opt = None,
        python_include: Tstr_opt = None,
    ) -> Self:
        path = Path(path)

        if not path.exists():
            raise ValueError(f"{path} does not exist")

        with open(path, "rb") as f:
            data = tomlkit.load(f)
        return cls(
            data=data, name=name, channels=channels, python_include=python_include
        )


# def _list_to_stream(values, stream=None):
#     value = "\n".join(values)
#     if isinstance(stream, (str, Path)):
#         with open(stream, "w") as f:
#             f.write(value)

#     elif stream is None:
#         return value

#     else:
#         stream.write(value)



# mypy: disable-error-code="no-untyped-def, no-untyped-call, assignment"
from pyproject2conda.cli import app

from click.testing import CliRunner

# from typer.testing import CliRunner

from textwrap import dedent

from pathlib import Path
import tempfile
import json

import sys

ROOT = Path(__file__).resolve().parent / "data"


def do_run(runner, command, *opts, filename=None, must_exist=False, **kwargs):
    if filename is None:
        filename = str(ROOT / "test-pyproject.toml")
    filename = Path(filename)
    if must_exist and not filename.exists():
        raise ValueError(f"filename {filename} does not exist")

    result = runner.invoke(app, [command, "-f", str(filename), *opts], **kwargs)

    return result


def check_result(result, expected):
    assert result.output == dedent(expected)


def test_list():
    runner = CliRunner()

    for cmd in ["l", "list"]:
        result = do_run(runner, cmd)
        expected = """\
        Extras:
        =======
        * test
        * dev-extras
        * dev
        * dist-pypi
        """
        check_result(result, expected)

    result = do_run(runner, "list", "-v")
    # TODO: the logger writes to output
    # assert result.output == f"filename: {ROOT / 'test-pyproject.toml'} [pyproject2conda - INFO]"
    check_result(result, expected)


def test_create():
    runner = CliRunner()

    # test unknown file

    result = do_run(runner, "yaml", filename="hello/there.toml")

    assert isinstance(result.exception, ValueError)

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    result = do_run(runner, "yaml")

    check_result(result, expected)

    cmd = " ".join([Path(sys.argv[0]).name] + sys.argv[1:])

    expected = f"""\
#
# This file is autogenerated by pyproject2conda
# with the following command:
#
#     $ {cmd}
#
# You should not manually edit this file.
# Instead edit the corresponding pyproject.toml file.
#
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    result = do_run(runner, "yaml", "--header")

    check_result(result, expected)

    # -p flag
    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8,<3.11
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    for opt in [("--python-include", "infer")]:
        for cmd in ["y", "yaml"]:
            result = do_run(runner, cmd, *opt)
            check_result(result, expected)

    # -p python=3.8
    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.8
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    for opts in [
        ("--python-include", "python=3.8"),
        ("--python", "3.8"),
        ("-p", "3.8"),
    ]:
        result = do_run(runner, "yaml", *opts)
        check_result(result, expected)

    # dev
    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - additional-thing
  - conda-matplotlib
  - pip
  - pip:
      - athing
    """

    for opt in ["-e", "--extra"]:
        result = do_run(runner, "yaml", opt, "dev", "--no-sort")
        check_result(result, expected)

    # test if add in "test" gives same answer
    result = do_run(runner, "yaml", "-e", "dev", "-e", "test", "--no-sort")
    check_result(result, expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - additional-thing
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - conda-matplotlib
  - pandas
  - pip
  - pip:
      - athing
    """

    for opt in ["-e", "--extra"]:
        result = do_run(runner, "yaml", opt, "dev")
        check_result(result, expected)

    # test if add in "test" gives same answer
    result = do_run(runner, "yaml", "-e", "dev", "-e", "test")
    check_result(result, expected)

    # different ordering
    expected = """\
channels:
  - conda-forge
dependencies:
  - conda-forge::cthing
  - bthing-conda
  - conda-forge::pytest
  - pandas
  - conda-matplotlib
  - additional-thing
  - pip
  - pip:
      - athing
    """

    for opt in ["-e", "--extra"]:
        result = do_run(
            runner,
            "yaml",
            opt,
            "dev",
            "--no-sort",
            filename=ROOT / "test-pyproject-reorder.toml",
            must_exist=True,
        )
        check_result(result, expected)

    # test if add in "test" gives same answer
    result = do_run(
        runner,
        "yaml",
        "-e",
        "dev",
        "-e",
        "test",
        "--no-sort",
        filename=ROOT / "test-pyproject-reorder.toml",
        must_exist=True,
    )
    check_result(result, expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - additional-thing
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - conda-matplotlib
  - pandas
  - pip
  - pip:
      - athing
    """

    for opt in ["-e", "--extra"]:
        result = do_run(
            runner,
            "yaml",
            opt,
            "dev",
            filename=ROOT / "test-pyproject-reorder.toml",
            must_exist=True,
        )
        check_result(result, expected)

    # test if add in "test" gives same answer
    result = do_run(
        runner,
        "yaml",
        "-e",
        "dev",
        "-e",
        "test",
        filename=ROOT / "test-pyproject-reorder.toml",
        must_exist=True,
    )
    check_result(result, expected)

    # override channel
    expected = """\
channels:
  - hello
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    for opt in ["-c", "--channel"]:
        result = do_run(runner, "yaml", opt, "hello")
        check_result(result, expected)

    expected = """\
channels:
  - hello
  - there
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    for opt in ["-c", "--channel"]:
        result = do_run(runner, "yaml", opt, "hello", opt, "there")
        check_result(result, expected)

    # test
    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - pip
  - pip:
      - athing
    """
    result = do_run(runner, "yaml", "-e", "test", "--no-sort")
    check_result(result, expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - pandas
  - pip
  - pip:
      - athing
    """
    result = do_run(runner, "yaml", "-e", "test")
    check_result(result, expected)

    # isolated
    expected = """\
channels:
  - conda-forge
dependencies:
  - setuptools
  - pip
  - pip:
      - build
    """
    for opt in ["-e", "--extra"]:
        result = do_run(runner, "yaml", opt, "dist-pypi", "--no-base")
        check_result(result, expected)


def test_requirements():
    runner = CliRunner()

    expected = """\
athing
bthing
cthing;python_version<'3.10'
    """

    for cmd in ["r", "requirements"]:
        result = do_run(runner, cmd)
        check_result(result, expected)

    expected = """\
athing
bthing
cthing;python_version<'3.10'
pandas
pytest
matplotlib
    """

    result = do_run(runner, "requirements", "-e", "dev", "--no-sort")
    check_result(result, expected)

    result = do_run(runner, "requirements", "-e", "dev", "-e", "test", "--no-sort")
    check_result(result, expected)

    expected = """\
athing
bthing
cthing;python_version<'3.10'
pandas
pytest
matplotlib
thing;python_version<'3.10'
other
    """

    result = do_run(
        runner,
        "requirements",
        "-e",
        "dev",
        "--no-sort",
        "-r",
        "thing;python_version<'3.10'",
        "-r",
        "other",
    )

    check_result(result, expected)

    expected = """\
athing
bthing
cthing;python_version<'3.10'
matplotlib
pandas
pytest
    """

    result = do_run(runner, "requirements", "-e", "dev")
    check_result(result, expected)

    result = do_run(runner, "requirements", "-e", "dev", "-e", "test")
    check_result(result, expected)

    expected = """\
athing
bthing
cthing;python_version<'3.10'
matplotlib
other
pandas
pytest
thing;python_version<'3.10'
    """

    result = do_run(
        runner,
        "requirements",
        "-e",
        "dev",
        "-r",
        "thing;python_version<'3.10'",
        "-r",
        "other",
    )

    check_result(result, expected)

    # allow whitespace:
    expected = """\
athing
bthing
cthing; python_version < '3.10'
matplotlib
other
pandas
pytest
thing; python_version < '3.10'
    """

    result = do_run(
        runner,
        "requirements",
        "-e",
        "dev",
        "-r",
        "thing; python_version < '3.10'",
        "-r",
        "other",
        "--no-remove-whitespace",
    )

    check_result(result, expected)

    expected = """\
setuptools
build
    """

    result = do_run(runner, "requirements", "-e", "dist-pypi", "--no-base", "--no-sort")
    check_result(result, expected)

    expected = """\
build
setuptools
    """

    result = do_run(runner, "requirements", "-e", "dist-pypi", "--no-base")
    check_result(result, expected)


def check_results_conda_req(path, expected):
    with open(path, "r") as f:
        result = f.read()

    assert result == dedent(expected)


def test_conda_requirements():
    runner = CliRunner()

    result = do_run(runner, "c", "hello.txt")

    assert isinstance(result.exception, ValueError)

    result = do_run(runner, "c", "--prefix", "hello", "a", "b")

    assert isinstance(result.exception, ValueError)

    # stdout
    result = do_run(runner, "c")

    expected = """\
#conda requirements
bthing-conda
conda-forge::cthing

#pip requirements
athing
    """

    assert result.output == dedent(expected)

    with tempfile.TemporaryDirectory() as d_tmp:
        d = Path(d_tmp)

        expected_conda = """\
bthing-conda
conda-forge::cthing
        """
        expected_pip = """\
athing
        """

        for cmd in ["c", "conda-requirements"]:
            do_run(runner, cmd, "--prefix", str(d / "hello-"), "--no-header")

            check_results_conda_req(d / "hello-conda.txt", expected_conda)
            check_results_conda_req(d / "hello-pip.txt", expected_pip)

        do_run(
            runner,
            "conda-requirements",
            str(d / "conda-output.txt"),
            str(d / "pip-output.txt"),
            "--no-header",
        )

        check_results_conda_req(d / "conda-output.txt", expected_conda)
        check_results_conda_req(d / "pip-output.txt", expected_pip)

        do_run(
            runner,
            "c",
            "--prefix",
            str(d / "hello-"),
            "--prepend-channel",
            "-c",
            "achannel",
            "--no-header",
        )

        expected_conda = """\
achannel::bthing-conda
conda-forge::cthing
        """

        check_results_conda_req(d / "hello-conda.txt", expected_conda)
        check_results_conda_req(d / "hello-pip.txt", expected_pip)


def test_json():
    runner = CliRunner()

    # stdout
    result = do_run(runner, "j")

    expected = """\
{"dependencies": ["bthing-conda", "conda-forge::cthing"], "pip": ["athing"], "channels": ["conda-forge"]}
    """

    assert result.output == dedent(expected)

    def check_results(path, expected):
        with open(path, "r") as f:
            result = json.load(f)

        assert result == expected

    with tempfile.TemporaryDirectory() as d_tmp:
        d = Path(d_tmp)

        expected = {  # type: ignore
            "dependencies": ["bthing-conda", "conda-forge::cthing"],
            "pip": ["athing"],
            "channels": ["conda-forge"],
        }

        do_run(runner, "json", "-o", str(d / "hello.json"))

        check_results(d / "hello.json", expected)

        expected = {  # type: ignore
            "dependencies": [
                "bthing-conda",
                "conda-forge::cthing",
                "pandas",
                "conda-forge::pytest",
                "additional-thing",
                "conda-matplotlib",
            ],
            "pip": ["athing"],
            "channels": ["conda-forge"],
        }

        do_run(runner, "json", "-o", str(d / "there.json"), "-e", "dev", "--no-sort")

        check_results(d / "there.json", expected)

        expected = {  # type: ignore
            "dependencies": [
                "additional-thing",
                "bthing-conda",
                "conda-forge::cthing",
                "conda-forge::pytest",
                "conda-matplotlib",
                "pandas",
            ],
            "pip": ["athing"],
            "channels": ["conda-forge"],
        }

        do_run(runner, "json", "-o", str(d / "there.json"), "-e", "dev")

        check_results(d / "there.json", expected)


def test_alias():
    runner = CliRunner()

    result = do_run(runner, "q")

    assert isinstance(result.exception, BaseException)


#     expected = """\
# Usage: app [OPTIONS] COMMAND [ARGS]...
# Try 'app --help' for help.

# Error: No such command 'q'.
#     """

#     check_result(result, expected)


def test_overwrite():
    runner = CliRunner(mix_stderr=True)

    with tempfile.TemporaryDirectory() as d_tmp:
        d = Path(d_tmp)

        path = d / "out.yaml"

        assert not path.exists()

        result = do_run(
            runner, "yaml", "-o", str(path), "-v", "-w", "force", catch_exceptions=False
        )
        # assert result.output.strip() == f"# Creating yaml {d_tmp}/out.yaml"

        orig_time = path.stat().st_mtime

        for cmd in ["check", "skip", "force"]:
            result = do_run(
                runner,
                "yaml",
                "-o",
                str(d / "out.yaml"),
                "-v",
                "-w",
                cmd,
                catch_exceptions=False,
            )

            if cmd == "force":
                assert path.stat().st_mtime > orig_time
            else:
                assert path.stat().st_mtime == orig_time

            # assert (
            #     result.output.strip()
            #     == f"# Skipping yaml {d_tmp}/out.yaml. Pass `-w force` to force recreate output"
            # )

        path = d / "out.txt"
        assert not path.exists()

        result = do_run(
            runner,
            "r",
            "-o",
            str(d / "out.txt"),
            "-v",
            "-w",
            "force",
            catch_exceptions=False,
        )
        orig_time = path.stat().st_mtime

        # assert result.output.strip() == f"# Creating requirements {d_tmp}/out.txt"

        for cmd in ["check", "skip", "force"]:
            result = do_run(
                runner, "r", "-o", str(path), "-v", "-w", cmd, catch_exceptions=False
            )

            if cmd == "force":
                assert path.stat().st_mtime > orig_time
            else:
                assert path.stat().st_mtime == orig_time
            # assert (
            #     result.output.strip()
            #     == f"# Skipping requirements {d_tmp}/out.txt. Pass `-w force` to force recreate output"
            # )



"""
Utility methods (:mod:`pyproject2conda.utils`)
==============================================
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Mapping, Sequence


# taken from https://github.com/conda/conda-lock/blob/main/conda_lock/common.py
def get_in(
    keys: Sequence[Any],
    nested_dict: Mapping[Any, Any],
    default: Any = None,
    factory: Callable[..., Any] | None = None,
) -> Any:
    """
    >>> foo = {'a': {'b': {'c': 1}}}
    >>> get_in(['a', 'b'], foo)
    {'c': 1}

    """
    import operator
    from functools import reduce

    try:
        return reduce(operator.getitem, keys, nested_dict)
    except (KeyError, IndexError, TypeError):
        if factory is not None:
            return factory()
        else:
            return default


# def compose_decorators(*decs: Dec[F]) -> Dec[F]:
#     from functools import reduce
#     def wrapper(func: F) -> F:
#         return reduce(lambda x, f: f(x), reversed(decs), func)
#     return wrapper


# def compose_decorators(*decs: Dec[F]) -> Dec[F]:
#     def wrapper(func: F) -> F:
#         for d in reversed(decs):
#             func = d(func)
#         return func

#     return wrapper


def parse_pythons(
    python_include: str | None,
    python_version: str | None,
    python: str | None,
) -> tuple[str | None, str | None]:
    if python:
        return f"python={python}", python
    else:
        return python_include, python_version


def update_target(
    target: str | Path | None,
    *deps: str | Path,
    overwrite: str = "check",
) -> bool:
    """Check if target is older than deps:"""

    if target is None:
        # No output file. always run.
        return True

    overwrite = overwrite.lower()
    target = Path(target)

    if overwrite == "force":
        update = True
    elif overwrite == "skip":
        update = not target.exists()

    elif overwrite == "check":
        if not target.exists():
            update = True
        else:
            # check times
            deps_filtered: list[Path] = []
            for d in map(Path, deps):
                if d.exists():
                    deps_filtered.append(d)

            target_time = target.stat().st_mtime

            update = any(target_time < dep.stat().st_mtime for dep in deps_filtered)
    else:
        raise ValueError(f"unknown option overwrite={overwrite}")

    return update


# * filename from template
def filename_from_template(
    template: str | None,
    python: str | None = None,
    python_version: str | None = None,
    env_name: str | None = None,
    ext: str | None = "yaml",
) -> str | None:
    """
    Create a filename from

    --python-include python=3.8 or --python-version 3.8 or --python 3.8
    py_version: 3.8
    py: 38

    env : name of environment
    """

    if template is None:
        return None

    kws: dict[str, str] = {}
    if python:
        py_version = python
    elif python_version:
        py_version = python_version
    # elif python_include is not None:
    #     import re

    #     m = re.match(".*?([0-9.]+)", python_include)
    #     py_version = m.group(1)  # type: ignore
    else:
        py_version = None

    if py_version:
        kws["py_version"] = py_version
        kws["py"] = py_version.replace(".", "")

    if env_name:
        kws["env"] = env_name

    if ext:
        template = template + f".{ext}"

    return template.format(**kws)



"""Unitilities to use with cog"""

from __future__ import annotations

import shlex
import subprocess
import textwrap
from functools import lru_cache


def wrap_command(cmd: str) -> str:
    x = textwrap.wrap(cmd.strip(), 80)

    if len(cmd) > 1:
        x[:-1] = [c + " \\" for c in x[:-1]]
        x[1:] = [" " * 4 + c for c in x[1:]]

    return "\n".join(x)


@lru_cache
def get_pyproject(path: str) -> list[str]:
    with open(path) as f:
        lines = [_.strip() for _ in f]
    return lines


def run_command(
    cmd: str,
    wrapper: str = "bash",
    include_cmd: bool = True,
    bounds: tuple[int | None, int | None] | None = None,
) -> None:
    args = shlex.split(cmd)
    output = subprocess.check_output(args)

    total = output.decode()

    if bounds is not None:
        x = total.split("\n")[bounds[0] : bounds[1]]
        if bounds[0] is not None:
            x = ["...\n"] + x
        if bounds[1] is not None:
            x = x + ["\n ...\n"]

        total = "\n".join(x)

    if include_cmd:
        cmd = wrap_command(cmd)

        total = f"$ {cmd}\n{total}"

    if wrapper:
        total = f"```{wrapper}\n" + total + "```\n"

    print(total)


def cat_lines(
    path: str = "tests/data/test-pyproject.toml",
    begin: str | int | None = None,
    end: str | int | None = None,
    begin_dot: bool = False,
    end_dot: bool = False,
) -> None:
    lines = get_pyproject(path)

    begin_dot = begin_dot or begin is not None
    end_dot = end_dot or end is not None

    if isinstance(begin, str):
        begin = lines.index(begin)
    if isinstance(end, str):
        end = lines.index(end)

    output = "\n".join(lines[slice(begin, end)])

    if begin_dot:
        output = "# ...\n" + output

    if end_dot:
        output = output + "\n# ..."

    output = "\n```toml\n" + output + "\n```\n"
    print(output)
